from matplotlib import pyplot as plt
import numpy as np


class SortingAccuracyWidget:
    def __init__(self, *, sorting_comparison, property_name=None):
        self._SC = sorting_comparison
        self._property_name = property_name

    def plot(self, mode='accuracy'):
        if mode == 'all':
            fig=plt.figure(figsize=(12,3))
            self._do_plot(ax=fig.add_subplot(1,3,1), show=False)
            self._do_plot_recall(ax=fig.add_subplot(1,3,2), show=False)
            self._do_plot_precision(ax=fig.add_subplot(1,3,3), show=False)
            plt.show()
        else:
            fig=plt.figure(figsize=(4,3))
            ax = fig.add_subplot(1,1,1)
            if mode == 'accuracy':
                self._do_plot(ax=ax)
            elif mode == 'recall' or mode == 'sensitivity':
                self._do_plot_recall(ax=ax)
            elif mode == 'precision':
                self._do_plot_precision(ax=ax)


    def _do_plot(self, ax=None, accuracy_thresh=.8, show=True):
        SC = self._SC
        units = SC.getSorting1().get_unit_ids()
        agreements = [SC.getAgreementFraction(unit) for unit in units]
        if self._property_name:
            xvals = SC.getSorting1().getUnitsProperty(unit_ids=units, property_name=self._property_name)
            ax.plot(xvals, agreements, '.')
            plt.xlabel(self._property_name)
        else:
            ax.plot(agreements, '.')
            plt.xticks([])

        plt.ylabel('Accuracy')
        plt.ylim(0,1)
        nUnits_above = np.sum(np.array(agreements) >= accuracy_thresh)
        plt.title('{} units > {} accuracy'.format(nUnits_above, accuracy_thresh))
        if show: plt.show()


    def _do_plot_recall(self, ax=None, recall_thresh=.8, show=True):
        SC = self._SC

        units = SC.getSorting1().get_unit_ids()
        recall = [1-SC.getFalsePositiveFraction(unit) for unit in units]
        if self._property_name:
            xvals = SC.getSorting1().getUnitsProperty(unit_ids=units, property_name=self._property_name)
            ax.plot(xvals, recall, '.')
            plt.xlabel(self._property_name)
        else:
            ax.plot(recall, '.')
            plt.xticks([])

        plt.ylabel('Recall')
        plt.ylim(0,1)
        nUnits_above = np.sum(np.array(recall) >= recall_thresh)
        plt.title('{} units > {} Recall'.format(nUnits_above, recall_thresh))
        if show: plt.show()


    def _do_plot_precision(self, ax=None, precision_thresh=.8, show=True):
        SC = self._SC

        units = SC.getSorting1().get_unit_ids()
        precision = [1-SC.getFalseNegativeFraction(unit) for unit in units]
        if self._property_name:
            xvals = SC.getSorting1().getUnitsProperty(unit_ids=units, property_name=self._property_name)
            ax.plot(xvals, precision, '.')
            plt.xlabel(self._property_name)
        else:
            ax.plot(precision, '.')
            plt.xticks([])

        plt.ylabel('Precision')
        plt.ylim(0,1)
        nUnits_above = np.sum(np.array(precision) >= precision_thresh)
        plt.title('{} units > {} precision'.format(nUnits_above, precision_thresh))
        if show: plt.show()



import numpy as np
import spikeextractors as se
from scipy.optimize import linear_sum_assignment


class SortingComparison():
    def __init__(self, sorting1, sorting2, delta_tp=20):
        self._sorting1 = sorting1
        self._sorting2 = sorting2
        self._delta_tp = delta_tp
        self._do_comparison()

    def getSorting1(self):
        return self._sorting1

    def getSorting2(self):
        return self._sorting2

    def getMappedSorting1(self):
        return MappedSortingExtractor(self._sorting1, self._unit_map12)

    def getMappedSorting2(self):
        return MappedSortingExtractor(self._sorting2, self._unit_map21)

    def getMatchingEventCount(self, unit1, unit2):
        if (unit1 is not None) and (unit2 is not None):
            a = self._matching_event_counts_12[unit1]
            if unit2 in a:
                return a[unit2]
            else:
                return 0
        # else if unit1 is not None:
        #    return self._missed_event_counts_1[unit1]
        # else if unit2 is not None:
        #    return self._missed_event_counts_2[unit2]
        else:
            raise Exception('getMatchingEventCount: unit1 and unit2 must not be None.')

    def _compute_agreement_score(self, num_matches, num1, num2):
        denom = num1 + num2 - num_matches
        if denom == 0:
            return 0
        return num_matches / denom

    def _compute_safe_frac(self, numer, denom):
        if denom == 0:
            return 0
        return numer / denom

    def getBestUnitMatch1(self, unit1):
        if unit1 in self._best_match_units_12:
            return self._best_match_units_12[unit1]
        else:
            return None

    def getBestUnitMatch2(self, unit2):
        if unit2 in self._best_match_units_21:
            return self._best_match_units_21[unit2]
        else:
            return None

    def getMatchingUnitList1(self, unit1):
        a = self._matching_event_counts_12[unit1]
        return list(a.keys())

    def getMatchingUnitList2(self, unit2):
        a = self._matching_event_counts_21[unit2]
        return list(a.keys())

    def getAgreementFraction(self, unit1=None, unit2=None):
        if (unit1 is not None) and (unit2 is None):
            unit2 = self.getBestUnitMatch1(unit1)
            if unit2 is None:
                return 0
        if (unit1 is None) and (unit2 is not None):
            unit1 = self.getBestUnitMatch2(unit2)
            if unit1 is None:
                return 0
        if (unit1 is None) and (unit2 is None):
            raise Exception('getAgreementFraction: at least one of unit1 and unit2 must not be None.')

        a = self._matching_event_counts_12[unit1]
        if unit2 not in a:
            return 0
        return self._compute_agreement_score(a[unit2], self._event_counts_1[unit1], self._event_counts_2[unit2])

    def getFalsePositiveFraction(self, unit1, unit2=None):
        if unit1 is None:
            raise Exception('getFalsePositiveFraction: unit1 must not be None')
        if unit2 is None:
            unit2 = self.getBestUnitMatch1(unit1)
            if unit2 is None:
                return 0

        a = self._matching_event_counts_12[unit1]
        if unit2 not in a:
            return 0
        return 1 - self._compute_safe_frac(a[unit2], self._event_counts_1[unit1])

    def getFalseNegativeFraction(self, unit1, unit2=None):
        if unit1 is None:
            raise Exception('getFalsePositiveFraction: unit1 must not be None')
        if unit2 is None:
            unit2 = self.getBestUnitMatch1(unit1)
            if unit2 is None:
                return 0

        a = self._matching_event_counts_12[unit1]
        if unit2 not in a:
            return 0
        return 1 - self._compute_safe_frac(a[unit2], self._event_counts_2[unit2])

    def _do_comparison(self):
        self._event_counts_1 = dict()
        self._event_counts_2 = dict()
        self._matching_event_counts_12 = dict()
        self._best_match_units_12 = dict()
        self._matching_event_counts_21 = dict()
        self._best_match_units_21 = dict()
        self._unit_map12 = dict()
        self._unit_map21 = dict()

        sorting1 = self._sorting1
        sorting2 = self._sorting2
        unit1_ids = sorting1.get_unit_ids()
        unit2_ids = sorting2.get_unit_ids()
        N1 = len(unit1_ids)
        N2 = len(unit2_ids)
        event_counts1 = np.zeros((N1)).astype(np.int64)
        for i1, u1 in enumerate(unit1_ids):
            times1 = sorting1.get_unit_spike_train(u1)
            event_counts1[i1] = len(times1)
            self._event_counts_1[u1] = len(times1)
        event_counts2 = np.zeros((N2)).astype(np.int64)
        for i2, u2 in enumerate(unit2_ids):
            times2 = sorting2.get_unit_spike_train(u2)
            event_counts2[i2] = len(times2)
            self._event_counts_2[u2] = len(times2)
        matching_event_counts = np.zeros((N1, N2)).astype(np.int64)
        scores = np.zeros((N1, N2))
        for i1, u1 in enumerate(unit1_ids):
            times1 = sorting1.get_unit_spike_train(u1)
            for i2, u2 in enumerate(unit2_ids):
                times2 = sorting2.get_unit_spike_train(u2)
                num_matches = count_matching_events(times1, times2, delta=self._delta_tp)
                matching_event_counts[i1, i2] = num_matches
                scores[i1, i2] = self._compute_agreement_score(num_matches, event_counts1[i1], event_counts2[i2])

        for i1, u1 in enumerate(unit1_ids):
            scores0 = scores[i1, :]
            self._matching_event_counts_12[u1] = dict()
            if np.max(scores0) > 0:
                inds0 = np.where(scores0 > 0)[0]
                for i2 in inds0:
                    self._matching_event_counts_12[u1][unit2_ids[i2]] = matching_event_counts[i1, i2]
                i2_best = np.argmax(scores0)
                self._best_match_units_12[u1] = unit2_ids[i2_best]
            else:
                self._best_match_units_12[u1] = None

        for i2, u2 in enumerate(unit2_ids):
            scores0 = scores[:, i2]
            self._matching_event_counts_21[u2] = dict()
            if np.max(scores0) > 0:
                inds0 = np.where(scores0 > 0)[0]
                for i1 in inds0:
                    self._matching_event_counts_21[u2][unit1_ids[i1]] = matching_event_counts[i1, i2]
                i1_best = np.argmax(scores0)
                self._best_match_units_21[u1] = unit1_ids[i1_best]
            else:
                self._best_match_units_21[u1] = None

        [inds1, inds2] = linear_sum_assignment(-scores)
        inds1 = list(inds1)
        inds2 = list(inds2)
        k2 = np.max(unit2_ids) + 1
        for i1, u1 in enumerate(unit1_ids):
            if i1 in inds1:
                aa = inds1.index(i1)
                i2 = inds2[aa]
                u2 = unit2_ids[i2]
                self._unit_map12[u1] = u2
            else:
                self._unit_map12[u1] = k2
                k2 = k2 + 1
        k1 = np.max(unit1_ids) + 1
        for i2, u2 in enumerate(unit2_ids):
            if i2 in inds2:
                aa = inds2.index(i2)
                i1 = inds1[aa]
                u1 = unit1_ids[i1]
                self._unit_map21[u2] = u1
            else:
                self._unit_map21[u2] = k1
                k1 = k1 + 1


class MappedSortingExtractor(se.SortingExtractor):
    def __init__(self, sorting, unit_map):
        se.SortingExtractor.__init__(self)
        self._sorting = sorting
        self._unit_map = unit_map
        self._reverse_map = dict()
        for key in unit_map:
            self._reverse_map[unit_map[key]] = key
        units = sorting.get_unit_ids()
        self._unit_ids = list(np.sort([self._unit_map[unit] for unit in units]))

    def get_unit_ids(self):
        return self._unit_ids

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        unit2 = self._reverse_map[unit_id]
        return self._sorting.get_unit_spike_train(unit2, start_frame=start_frame, end_frame=end_frame)


def count_matching_events(times1, times2, delta=20):
    times_concat = np.concatenate((times1, times2))
    membership = np.concatenate((np.ones(times1.shape) * 1, np.ones(times2.shape) * 2))
    indices = times_concat.argsort()
    times_concat_sorted = times_concat[indices]
    membership_sorted = membership[indices]
    diffs = times_concat_sorted[1:] - times_concat_sorted[:-1]
    inds = np.where((diffs <= delta) & (membership_sorted[0:-1] != membership_sorted[1:]))[0]
    if (len(inds) == 0):
        return 0
    inds2 = np.where(inds[:-1] + 1 != inds[1:])[0]
    return len(inds2) + 1



import vdomr as vd


class RecordingSummaryView(vd.Component):
    def __init__(self, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)

    @staticmethod
    def prepareView(context, opts):
        context.initialize()

    def setSize(self, size):
        self._size = size

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Recording summary'

    def render(self):
        rows = []
        rows.append(vd.tr(
            vd.th('Study'), vd.td(self._context.studyName())
        ))
        rows.append(vd.tr(
            vd.th('Recording'), vd.td(self._context.recordingName())
        ))
        rows.append(vd.tr(
            vd.th('Directory'), vd.td(self._context.recordingDirectory())
        ))
        RX = self._context.recordingExtractor()
        rows.append(vd.tr(
            vd.th('Num. channels'), vd.td('{}'.format(len(RX.get_channel_ids())))
        ))
        rows.append(vd.tr(
            vd.th('Samplerate'), vd.td('{}'.format(RX.get_sampling_frequency()))
        ))
        a = RX.get_num_frames() / RX.get_sampling_frequency()
        rows.append(vd.tr(
            vd.th('Duration (s)'), vd.td('{}'.format(a))
        ))

        a = ', '.join(self._context.sortingResultNames())
        rows.append(vd.tr(
            vd.th('Sorting results'), vd.td('{}'.format(a))
        ))

        sc_true = self._context.trueSortingContext()
        if sc_true:
            SX_true = sc_true.sortingExtractor()
            true_unit_ids = SX_true.get_unit_ids()
            rows.append(vd.tr(
                vd.th('Num. true units'), vd.td('{}'.format(len(true_unit_ids)))
            ))

        table = vd.table(rows, style={
                         'text-align': 'left', 'width': 'auto', 'font-size': '13px'}, class_='table')

        return vd.div(
            vd.h2('{}/{}'.format(self._context.studyName(), self._context.recordingName())),
            table
        )



import pandas as pd
from IPython.display import HTML


class SortingComparisonTable():
    def __init__(self, comparison, reference=1, unit_properties=[]):
        self._comparison = comparison
        self._unit_properties = unit_properties
        self._reference = reference
        for i in range(len(self._unit_properties)):
            prop = self._unit_properties[i]
            if type(prop) != dict:
                self._unit_properties[i] = {'name': prop}

    def getDataframe(self):
        SC = self._comparison
        rows = []
        if self._reference == 1:
            for _, unit1 in enumerate(SC.getSorting1().get_unit_ids()):
                unit2 = SC.getBestUnitMatch1(unit1)
                row0 = {
                    'unit_id': unit1,
                    'accuracy': SC.getAgreementFraction(unit1, unit2),
                    'best_unit': unit2,
                    'matched_unit': SC.getMappedSorting1().getMappedUnitIds(unit1),
                    'num_matches': SC.getMatchingEventCount(unit1, unit2),
                    'f_n': SC.getFalseNegativeFraction(unit1),
                    'f_p': SC.getFalsePositiveFraction(unit1),
                }
                for prop in self._unit_properties:
                    pname = prop['name']
                    row0[pname] = SC.getSorting1().get_unit_property(unit_id=int(unit1), property_name=pname)
                rows.append(row0)
        elif self._reference == 2:
            for u_1, unit1 in enumerate(SC.getSorting2().get_unit_ids()):
                unit2 = SC.getBestUnitMatch2(unit1)
                row0 = {
                    'unit_id': unit1,
                    'accuracy': SC.getAgreementFraction(unit2, unit1),
                    'best_unit': unit2,
                    'matched_unit': SC.getMappedSorting2().getMappedUnitIds(unit1),
                    'num_matches': SC.getMatchingEventCount(unit2, unit1),
                    'f_n': SC.getFalseNegativeFraction(unit1),
                    'f_p': SC.getFalsePositiveFraction(unit1),
                }
                for prop in self._unit_properties:
                    pname = prop['name']
                    row0[pname] = SC.getSorting2().get_unit_property(unit_id=int(unit1), property_name=pname)
                rows.append(row0)

        df = pd.DataFrame(rows)
        fields = ['unit_id']
        fields = fields + ['accuracy', 'best_unit', 'matched_unit', 'f_n', 'f_p', 'num_matches']
        for prop in self._unit_properties:
            pname = prop['name']
            fields.append(pname)
        df = df[fields]
        df['accuracy'] = df['accuracy'].map('{:,.2f}'.format)
        # df['Best match'] = df['Accuracy'].map('{:,.2f}'.format)
        df['f_n'] = df['f_n'].map('{:,.2f}'.format)
        df['f_p'] = df['f_p'].map('{:,.2f}'.format)
        return df

    def display(self):
        df = self.getDataframe()
        display(HTML(df.to_html(index=False)))



# %% Change working directory from the workspace root to the ipynb file location. Turn this addition off with the DataScience.changeDirOnImportExport setting
import os
try:
    os.chdir(os.path.join(os.getcwd(), 'working/spikeforest_test1'))
    print(os.getcwd())
except:
    pass

# %%
import spikeforest_analysis as sa
from spikeforest_analysis.compare_sortings_with_truth import GenSortingComparisonTable
#from mountaintools import client as ca
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor, example_datasets
import os
from spikesorters import YASS
from spikeforest import spikewidgets as sw

# %%SortingComparisonyass_test1/recording/raw.mda
tmpdir = 'yass_test1'
if not os.path.isdir(tmpdir):
    os.mkdir(tmpdir)
rx, sx = example_datasets.yass_example(set_id=1)

# %%
firings_true = tmpdir+'/recording/firings_true.mda'
recording_path = tmpdir+'/recording'
SFMdaRecordingExtractor.write_recording(
    recording=rx, save_path=recording_path)
SFMdaSortingExtractor.write_sorting(
    sorting=sx, save_path=firings_true)

YASS.execute(
    recording_dir=tmpdir+'/recording',
    firings_out=tmpdir+'/firings_out.mda',
    detect_sign=-1,
    adjacency_radius=50,
    _container=None,
    _force_run=True,
    _keep_temp_files=True
)
firings_out = tmpdir+'/firings_out.mda'
assert os.path.exists(firings_out)

# %%
print('recording: {}'.format(recording_path))
print('firings_out: {}'.format(firings_out))
print('firings_true: {}'.format(firings_true))

# %%
GenSortingComparisonTable.execute(
    firings=firings_out,
    firings_true=firings_true,
    units_true=[],
    json_out=os.path.join(tmpdir, 'out.json'),
    html_out=os.path.join(tmpdir, 'out.html'),
    _container=None
)


# %%



import numpy as np


def synthesize_timeseries(*, sorting, waveforms, noise_level=1, samplerate=30000, duration=60, waveform_upsamplefac=13):
    num_timepoints = np.int64(samplerate * duration)
    waveform_upsamplefac = int(waveform_upsamplefac)
    W = waveforms

    M, TT, _ = W.shape[0], W.shape[1], W.shape[2]
    T = int(TT / waveform_upsamplefac)
    Tmid = int(np.ceil((T + 1) / 2 - 1))

    N = num_timepoints

    X = np.random.randn(M, N) * noise_level

    unit_ids = sorting.get_unit_ids()
    for k0 in unit_ids:
        waveform0 = waveforms[:, :, k0 - 1]
        times0 = sorting.get_unit_spike_train(unit_id=k0)
        for t0 in times0:
            amp0 = 1
            frac_offset = int(np.floor((t0 - np.floor(t0)) * waveform_upsamplefac))
            tstart = np.int64(np.floor(t0)) - Tmid
            if (0 <= tstart) and (tstart + T <= N):
                X[:, tstart:tstart + T] = X[:, tstart:tstart + T] + waveform0[:, frac_offset::waveform_upsamplefac] * amp0

    return X



import sys
from mountaintools import client as mt
from .core import ForestViewMainWindow
from .spikeforestcontext import SpikeForestContext
from .analysiscontext import AnalysisContext
from .spikefrontcontext import SpikeFrontContext
import os


def forestview(path, *, mode='spikeforest'):
    if mode == 'spikeforest':
        context = _load_spikeforest_context(path)
    elif mode == 'analysis':
        context = _load_analysis_context(path)
    elif mode == 'spike-front':
        context = _load_spikefront_context(path)
    else:
        raise Exception('Invalid mode: ' + mode)

    if not context:
        raise Exception('Unable to create context.')

    W = ForestViewMainWindow(context=context)

    return W


def _load_spikeforest_context(path):
    if mt.isFile(path):
        obj = mt.loadObject(path=path)
        if not obj:
            print('Unable to load file: ' + path, file=sys.stderr)
            return None
    else:
        obj = _make_obj_from_dir(path)
        if not obj:
            print('Unable to make object from path: ' + path)
            return None
    context = SpikeForestContext(
        studies=obj.get('studies', []),
        recordings=obj.get('recordings', []),
        sorting_results=obj.get('sorting_results', []),
        aggregated_sorting_results=obj.get('aggregated_sorting_results', None)
    )
    return context


def _load_analysis_context(path):
    obj = mt.loadObject(path=path)
    if not obj:
        print('Unable to load file: ' + path, file=sys.stderr)
        return None
    context = AnalysisContext(
        obj=obj
    )
    return context


def _load_spikefront_context(path):
    obj = mt.loadObject(path=path)
    if not obj:
        print('Unable to load file: ' + path, file=sys.stderr)
        return None
    context = SpikeFrontContext(
        StudySets=obj.get("StudySets", []),
        Recordings=obj.get("Recordings", []),
        TrueUnits=obj.get("TrueUnits", []),
        UnitResults=obj.get("UnitResults", []),
        SortingResults=obj.get("SortingResults", []),
        Sorters=obj.get("Sorters", []),
        Studies=obj.get("Studies", []),
        Algorithms=obj.get("Algorithms", []),
        StudyAnalysisResults=obj.get("StudyAnalysisResults", [])
    )
    return context


def _make_obj_from_dir(path):
    studies = []
    recordings = []
    study_name = os.path.basename(path)
    studies.append(dict(
        name=study_name,
        description='Loaded from ' + path
    ))
    dd = mt.readDir(path)
    if not dd:
        print('Unable to read directory: ' + path)
        return None
    if 'raw.mda' in dd['files']:
        recordings.append(_make_recording_obj_from_dir(
            path=path,
            study_name=study_name,
            name='recording'
        ))
    else:
        for dname in dd['dirs'].keys():
            recordings.append(_make_recording_obj_from_dir(
                path=path + '/' + dname,
                study_name=study_name,
                name=dname
            ))
    return dict(
        studies=studies,
        recordings=recordings
    )


def _make_recording_obj_from_dir(*, path, study_name, name):
    ret = dict(
        study=study_name,
        name=name,
        directory=path
    )
    if mt.computeFileSha1(path + '/firings_true.mda'):
        ret['firings_true'] = path + '/firings_true.mda'
    intra_raw_fname = path + '/raw_true.mda'
    if mt.computeFileSha1(intra_raw_fname):
        ret['intra_recording'] = dict(
            study=study_name,
            name=name + '--intra',
            directory=path,
            raw_fname='raw_true.mda',
            params_fname='params_true.json'
        )
    return ret



import vdomr as vd
import os

source_path = os.path.dirname(os.path.realpath(__file__))


class TabBar(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._height = 46  # hard-coded to match css of ChromeTabs
        self._current_tab_id = None
        self._current_tab_changed_handlers = []
        self._tab_removed_handlers = []
        self._tab_labels = dict()

        vd.devel.loadJavascript(path=source_path + '/draggabilly.js')
        vd.devel.loadCss(path=source_path + '/chrome-tabs/css/chrome-tabs.css')
        vd.devel.loadCss(path=source_path + '/chrome-tabs/css/chrome-tabs-dark-theme.css')
        vd.devel.loadCss(css=_CSS)
        vd.devel.loadJavascript(path=source_path + '/chrome-tabs/js/chrome-tabs.js')

    def height(self):
        return self._height

    def onCurrentTabChanged(self, handler):
        self._current_tab_changed_handlers.append(handler)

    def onTabRemoved(self, handler):
        self._tab_removed_handlers.append(handler)

    def setCurrentTab(self, id):
        js = """
        let chromeTabs = window.chrome_tab_widgets['{component_id}'];
        let el = document.querySelector('#div-{component_id}');
        let tabEl = $(el).find('[data-tab-id={tab_id}]');
        if (tabEl)
            chromeTabs.setCurrentTab(tabEl[0]);
        """
        js = js.replace('{tab_id}', id)
        js = js.replace('{component_id}', self.componentId())
        self.executeJavascript(js=js)

    def currentTabId(self):
        return self._current_tab_id

    def addTab(self, id, label):
        js = """
        let chromeTabs = window.chrome_tab_widgets['{component_id}'];
        chromeTabs.addTab({
            id: '{id}',
            title: 'initializing...',
            favicon: false
        });
        """
        js = js.replace('{id}', id)
        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{label}', label)
        self.executeJavascript(js=js)
        self._on_active_tab_changed(id)
        self.setTabLabel(id, label)
        pass

    def setTabLabel(self, id, label):
        if id in self._tab_labels:
            if label == self._tab_labels[id]:
                return
        self._tab_labels[id] = label
        js = """
        let el = document.querySelector('#div-{component_id}');
        let tabEl = $(el).find('[data-tab-id={tab_id}]');
        $(tabEl).find('.chrome-tab-title').html('{label}');
        """
        js = js.replace('{tab_id}', id)
        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{label}', label)
        self.executeJavascript(js)

    def render(self):
        div = vd.div(vd.div(class_="chrome-tabs-content"), vd.div(class_="chrome-tabs-bottom-bar"), class_='chrome-tabs', id='div-' + self.componentId())
        return div

    def _on_active_tab_changed(self, id):
        if self._current_tab_id == id:
            return
        self._current_tab_id = id
        for handler in self._current_tab_changed_handlers:
            handler()

    def _on_tab_removed(self, id):
        for handler in self._tab_removed_handlers:
            handler(id)
        if id == self._current_tab_id:
            self._on_active_tab_changed(None)

    def postRenderScript(self):
        js = """
        let chromeTabs = new ChromeTabs()
        window.chrome_tab_widgets=window.chrome_tab_widgets||{};
        window.chrome_tab_widgets['{component_id}']=chromeTabs;

        let el = document.querySelector('#div-{component_id}');
        chromeTabs.init(el);
        
        el.addEventListener('activeTabChange', function(evt) {
            let id = evt.detail.tabEl.getAttribute('data-tab-id');
            {on_active_tab_changed}([],{id:id});

        });
        el.addEventListener('tabRemove', function(evt) {
            let id = evt.detail.tabEl.getAttribute('data-tab-id');
            {on_tab_removed}([],{id:id});

        });
        """
        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{on_active_tab_changed}', vd.create_callback(self._on_active_tab_changed))
        js = js.replace('{on_tab_removed}', vd.create_callback(self._on_tab_removed))
        return js

_CSS = """
.chrome-tabs .chrome-tab {
width: 258px
}

.chrome-tabs .chrome-tab:nth-child(1) {
transform: translate3d(0px, 0, 0)
}

.chrome-tabs .chrome-tab:nth-child(2) {
transform: translate3d(239px, 0, 0)
}
"""



import uuid
import vdomr as vd
import os
import json

source_path = os.path.dirname(os.path.realpath(__file__))


class ElectrodeGeometryWidget(vd.Component):
    def __init__(self, *, recording):
        vd.Component.__init__(self)
        self._recording = recording
        self._div_id = 'ElectrodeGeometryWidget-' + str(uuid.uuid4())

        vd.devel.loadBootstrap()
        vd.devel.loadCss(url='https://stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css')
        vd.devel.loadJavascript(path=source_path + '/dist/main.js')
        vd.devel.loadJavascript(path=source_path + '/dist/d3.v5.min.js')
        vd.devel.loadJavascript(path=source_path + '/dist/jquery-3.3.1.min.js')

        js = "window.sfdata=window.sfdata||{}; window.sfdata['electrode_geometry_{div_id}']={obj}"
        js = self._div_id.join(js.split('{div_id}'))
        obj = self._get_geom()
        js = json.dumps(obj).join(js.split('{obj}'))
        vd.devel.loadJavascript(js=js)

    def _get_geom(self):
        RX = self._recording
        electrodes = []
        for ch in RX.get_channel_ids():
            location = RX.get_channel_property(channel_id=ch, property_name='location')
            electrodes.append(
                dict(
                    channel_id=ch,
                    location=list(location)
                )
            )
        obj = dict(electrodes=electrodes)
        return obj

    def render(self):
        div = vd.div(id=self._div_id)
        js = """
        let W=new window.GeomWidget();
        let X=window.sfdata['electrode_geometry_{z_id}']
        console.log(X);
        W.setGeometry(X);
        //W.setSize(300,300)
        $('#{div_id}').empty();
        $('#{div_id}').append(W.div());
        W.update();
        """
        js = self._div_id.join(js.split('{div_id}'))
        vd.devel.loadJavascript(js=js, delay=1)
        return div



import numpy as np
from matplotlib import pyplot as plt
import vdomr as vd
import time
from spikeforest import spikewidgets as sw


class CorrelogramsWidget(vd.Component):
    def __init__(self, *, sorting, samplerate):
        vd.Component.__init__(self)
        self._widgets = [
            CorrelogramWidget(
                sorting=sorting,
                samplerate=samplerate,
                unit1_id=id,
                unit2_id=id
            )
            for id in sorting.get_unit_ids()
        ]
        vd.devel.loadBootstrap()

    def setSelectedUnitIds(self, ids):
        ids = set(ids)
        for W in self._widgets:
            W.setSelected(W.unit1Id() in ids)

    def render(self):
        box_style = dict(float='left')
        boxes = [
            vd.div(W, style=box_style)
            for W in self._widgets
        ]
        div = vd.div(boxes)
        return div


class CorrelogramWidget(vd.Component):
    def __init__(self, *, sorting, samplerate, unit1_id, unit2_id):
        vd.Component.__init__(self)
        self._plot = _CorrelogramPlot(
            sorting=sorting,
            samplerate=samplerate,
            unit1_id=unit1_id,
            unit2_id=unit2_id
        )
        self._plot_div = vd.components.LazyDiv(self._plot)
        self._unit1_id = unit1_id
        self._unit2_id = unit2_id
        self._selected = False

    def setSelected(self, val):
        if self._selected == val:
            return
        self._selected = val
        self.refresh()

    def unit1Id(self):
        return self._unit1_id

    def unit2Id(self):
        return self._unit1_id

    def render(self):
        style0 = {'border': 'solid 1px black', 'margin': '5px'}
        style1 = {}
        if self._selected:
            style1['background-color'] = 'yellow'
        if self._unit1_id == self._unit2_id:
            title0 = 'Unit {}'.format(self._unit1_id)
        else:
            title0 = '{} / {}'.format(self._unit1_id, self._unit2_id)
        return vd.div(
            vd.p(title0, style={'text-align': 'center'}),
            vd.div(self._plot_div, style=style0),
            style=style1
        )


class _CorrelogramPlot(vd.components.Pyplot):
    def __init__(self, *, sorting, samplerate, unit1_id, unit2_id):
        vd.components.Pyplot.__init__(self)
        self._sorting = sorting
        self._samplerate = samplerate
        self._unit1_id = unit1_id
        self._unit2_id = unit2_id

    def plot(self):
        plot_correlogram(
            sorting=self._sorting,
            samplerate=self._samplerate,
            unit1_id=self._unit1_id,
            unit2_id=self._unit2_id
        )


def _plot_correlogram_helper(*, bin_counts, bin_edges, title=''):
    wid = (bin_edges[1] - bin_edges[0]) * 1000
    plt.bar(x=bin_edges[0:-1] * 1000, height=bin_counts, width=wid, color='gray', align='edge')
    plt.xlabel('dt (msec)')
    plt.gca().get_yaxis().set_ticks([])
    plt.gca().get_xaxis().set_ticks([])
    plt.gca().get_yaxis().set_ticks([])
    if title:
        plt.title(title, color='gray')


def compute_autocorrelogram(times, *, max_dt_tp, bin_size_tp, max_samples=None):
    num_bins_left = int(max_dt_tp / bin_size_tp)  # number of bins to the left of the origin
    L = len(times)  # number of events
    times2 = np.sort(times)  # the sorted times
    step = 1  # This is the index step between an event and the next one to compare
    candidate_inds = np.arange(L)  # These are the events we are going to consider
    if max_samples is not None:
        if len(candidate_inds) > max_samples:
            candidate_inds = np.random.choice(candidate_inds, size=max_samples, replace=False)
    vals_list = []  # A list of all offsets we have accumulated
    while True:
        candidate_inds = candidate_inds[
            candidate_inds + step < L]  # we only consider events that are within workable range
        candidate_inds = candidate_inds[times2[candidate_inds + step] - times2[
            candidate_inds] <= max_dt_tp]  # we only consider event-pairs that are within max_dt_tp apart
        if len(candidate_inds) > 0:  # if we have some events to consider
            vals = times2[candidate_inds + step] - times2[candidate_inds]
            vals_list.append(vals)  # add to the autocorrelogram
            vals_list.append(-vals)  # keep it symmetric
        else:
            break  # no more to consider
        step += 1
    if len(vals_list) > 0:  # concatenate all the values
        all_vals = np.concatenate(vals_list)
    else:
        all_vals = np.array([])
    aa = np.arange(-num_bins_left, num_bins_left + 1) * bin_size_tp
    all_vals = np.sign(all_vals) * (np.abs(
        all_vals) - bin_size_tp * 0.00001)  # a trick to make the histogram symmetric due to differences in rounding for positive and negative, i suppose
    bin_counts, bin_edges = np.histogram(all_vals, bins=aa)
    return (bin_counts, bin_edges)


def plot_correlogram(*, sorting, samplerate, unit1_id, unit2_id, title=''):
    if unit1_id != unit2_id:
        raise Exception('This case not supported yet.')
    times = sorting.get_unit_spike_train(unit_id=unit1_id)
    max_dt_msec = 50
    bin_size_msec = 2
    max_dt_tp = max_dt_msec * samplerate / 1000
    bin_size_tp = bin_size_msec * samplerate / 1000
    (bin_counts, bin_edges) = compute_autocorrelogram(times, max_dt_tp=max_dt_tp, bin_size_tp=bin_size_tp)
    _plot_correlogram_helper(bin_counts=bin_counts, bin_edges=bin_edges, title=title)



import vdomr as vd
from .tabbar import TabBar


class HighlightBox(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._highlight = True

    def setHighlight(self, val):
        if self._highlight == val:
            return
        self._highlight = val
        self.refresh()

    def render(self):
        style0 = dict(width='100%', height='100%', position='absolute')
        style0['z-index'] = '-1'
        if self._highlight:
            style0['border'] = 'solid 2px black'
        return vd.div(style=style0)


class ViewHolder(vd.Component):
    def __init__(self, child, *, name=''):
        vd.Component.__init__(self)
        self._child = child
        self._name = name

    def name(self):
        return self._name

    def child(self):
        return self._child

    def render(self):
        return vd.div(self._child)


class Container(vd.Component):
    def __init__(self, *args, position, size, position_mode='absolute', style=dict()):
        vd.Component.__init__(self)
        self._children = list(args)
        self._position = position
        self._size = size
        self._position_mode = position_mode
        self._style = style

    def render(self):
        style = self._style
        style['position'] = self._position_mode
        style['width'] = '{}px'.format(self._size[0])
        style['height'] = '{}px'.format(self._size[1])
        style['left'] = '{}px'.format(self._position[0])
        style['top'] = '{}px'.format(self._position[1])
        return vd.div(
            self._children,
            style=style
        )


class ViewContainerContent(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._child = None

    def setChild(self, child):
        self._child = child
        self.refresh()

    def render(self):
        if not self._child:
            return vd.div('No content')
        return self._child


class ViewContainer(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._views = dict()
        self._fiew_holders = dict()
        self._content = ViewContainerContent()
        self._last_id_num = 0
        self._size = (0, 0)
        self._tab_bar = TabBar()
        self._tab_bar.onCurrentTabChanged(self._on_current_tab_changed)
        self._tab_bar.onTabRemoved(self._on_tab_removed)
        self._click_handlers = []
        self._highlight_box = HighlightBox()
        self._poll()

    def onClick(self, handler):
        self._click_handlers.append(handler)

    def setSize(self, size):
        self._size = size
        for view in self._views.values():
            self._update_view_size(view)

    def addView(self, view, *, name=''):
        id_num = self._last_id_num + 1
        self._last_id_num = self._last_id_num + 1
        view_id = 'view_{}'.format(id_num)

        self._update_view_size(view)
        holder = ViewHolder(view, name=name)
        self._views[view_id] = view
        self._fiew_holders[view_id] = holder
        self._tab_bar.addTab(view_id, view.tabLabel())

        # self.refresh()
    def findView(self, *, name):
        for vf in self._fiew_holders.values():
            if vf.name() == name:
                return vf.child()
        return None

    def setCurrentView(self, view):
        for id, v in self._views.items():
            if v == view:
                self._tab_bar.setCurrentTab(id=id)
                return

    def setHighlight(self, val):
        self._highlight_box.setHighlight(val)

    def _update_view_size(self, view):
        view.setSize((self._size[0] - self._tab_bar.height() - 10, self._size[1] - self._tab_bar.height() - 10))

    def _on_current_tab_changed(self):
        id = self._tab_bar.currentTabId()
        if id:
            holder = self._fiew_holders[id]
            self._content.setChild(holder)
        else:
            self._content.setChild(None)

    def _on_tab_removed(self, id):
        if id in self._views:
            view = self._views[id]
            del self._views[id]
            del self._fiew_holders[id]
            if hasattr(view, 'cleanup'):
                (getattr(view, 'cleanup'))()

    def currentView(self):
        id = self._tab_bar.currentTabId()
        if id is None:
            return None
        if id not in self._views:
            return None
        return self._views[id]

    def _on_click(self):
        for handler in self._click_handlers:
            handler()

    def _poll(self):
        for id, v in self._views.items():
            label = v.tabLabel()
            self._tab_bar.setTabLabel(id, label)
            if hasattr(v, 'updateTitle'):
                v.updateTitle()
        vd.set_timeout(self._poll, 2)

    def render(self):
        style_outer = dict(width='100%', height='100%', position='absolute')
        style_content = dict(left='5px', right='5px', top='{}px'.format(5 + self._tab_bar.height()), bottom='5px', position='absolute')
        onclick = self._on_click
        return vd.div(self._highlight_box, vd.div(self._content, id='content-' + self.componentId(), style=style_content), self._tab_bar, style=style_outer, onclick=onclick)



from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor, EfficientAccessRecordingExtractor
from copy import deepcopy
import spikeforestwidgets as SFW
from mountaintools import client as mt
from .bandpass_filter import bandpass_filter


class RecordingContext():
    def __init__(self, recording_object, *, download=True, create_earx=True, precompute_multiscale=True):
        from .sortingresultcontext import SortingResultContext  # avoid cyclic dependency

        self._signal_handlers = dict()
        self._any_state_change_handlers = []
        self._recording_object = recording_object
        self._download = download

        self._sorting_result_contexts = dict()

        if 'firings_true' in recording_object:
            sc_true = SortingResultContext(sorting_result_object=dict(firings=recording_object['firings_true']), recording_context=self)
            self._true_sorting_context = sc_true
            sc_true.onAnyStateChanged(self._trigger_any_state_change_handlers)
        else:
            self._true_sorting_context = None

        if self._recording_object.get('intra_recording', None):
            self._intra_recording_context = RecordingContext(recording_object=self._recording_object['intra_recording'], download=download, create_earx=create_earx, precompute_multiscale=precompute_multiscale)
        else:
            self._intra_recording_context = None

        self._state = dict(
            current_timepoint=None,
            selected_time_range=None,
            current_channel=None,
            current_sorting_result=None
        )

        self._initialized = False

    def initialize(self):
        if self._initialized:
            return
        self._initialized = True

        print('******** FORESTVIEW: Initializing recording context')
        self._recording_object = self._recording_object
        if self._download:
            print('******** FORESTVIEW: Downloading recording file if needed...')
        recdir = self._recording_object['directory']
        raw_fname = self._recording_object.get('raw_fname', 'raw.mda')
        params_fname = self._recording_object.get('params_fname', 'params.json')
        self._rx = SFMdaRecordingExtractor(dataset_directory=recdir, download=self._download, raw_fname=raw_fname, params_fname=params_fname)
        self._rx = bandpass_filter(self._rx)

        if self._true_sorting_context:
            self._true_sorting_context.initialize()

        # firings_true_path = recdir + '/firings_true.mda'
        # self._sx_true = None
        # if mt.computeFileSha1(path=firings_true_path):
        #     print('******** FORESTVIEW: Downloading true firings file if needed...')
        #     if not mt.realizeFile(firings_true_path):
        #         print('Warning: unable to realize true firings file: '+firings_true_path)
        #     else:
        #         self._sx_true = SFMdaSortingExtractor(firings_file = firings_true_path)

        if self._intra_recording_context:
            self._intra_recording_context.initialize()

        print('******** FORESTVIEW: Done initializing recording context')

    def sortingResultNames(self):
        return sorted(list(self._sorting_result_contexts.keys()))

    def sortingResultContext(self, name):
        return self._sorting_result_contexts[name]

    def addSortingResult(self, sorting_result_object):
        from .sortingresultcontext import SortingResultContext  # avoid cyclic dependency
        sc = SortingResultContext(sorting_result_object=sorting_result_object, recording_context=self)
        sc.onAnyStateChanged(self._trigger_any_state_change_handlers)
        self._sorting_result_contexts[sorting_result_object['sorter']['name']] = sc

    def recordingObject(self):
        return deepcopy(self._recording_object)

    def studyName(self):
        return self._recording_object.get('study', '')

    def recordingName(self):
        return self._recording_object.get('name', '')

    def recordingLabel(self):
        return '{}/{}'.format(self.studyName(), self.recordingName())

    def recordingDirectory(self):
        return self._recording_object.get('directory', '')

    def recordingExtractor(self):
        return self._rx

    def trueSortingContext(self):
        return self._true_sorting_context

    def intraRecordingContext(self):
        return self._intra_recording_context

    def hasIntraRecording(self):
        return self._intra_recording_context is not None

    def onAnyStateChanged(self, handler):
        for key in self._state.keys():
            self._register_state_change_handler(key, handler)
        self._any_state_change_handlers.append(handler)

    def stateObject(self):
        ret = deepcopy(self._state)
        sc = self.trueSortingContext()
        if sc:
            ret['true_sorting'] = sc.stateObject(include_recording=False)
        else:
            ret['true_sorting'] = None
        return ret

    # current channel
    def currentChannel(self):
        return self._get_state_value('current_channel')

    def setCurrentChannel(self, ch):
        if ch is not None:
            ch = int(ch)
        self._set_state_value('current_channel', ch)

    def onCurrentChannelChanged(self, handler):
        self._register_state_change_handler('current_channel', handler)

    # current timepoint
    def setCurrentTimepoint(self, tp):
        tp = int(tp)
        self._set_state_value('current_timepoint', tp)

    def currentTimepoint(self):
        return self._get_state_value('current_timepoint')

    def onCurrentTimepointChanged(self, handler):
        self._register_state_change_handler('current_timepoint', handler)

    # current time range
    def setCurrentTimeRange(self, range):
        range = (int(range[0]), int(range[1]))
        self._set_state_value('current_time_range', range)

    def currentTimeRange(self):
        return self._get_state_value('current_time_range')

    def onCurrentTimeRangeChanged(self, handler):
        self._register_state_change_handler('current_time_range', handler)

    # current sorting result
    def currentSortingResult(self):
        return self._get_state_value('current_sorting_result')

    def setCurrentSortingResult(self, srname):
        self._set_state_value('current_sorting_result', srname)

    def onCurrentSortingResultChanged(self, handler):
        self._register_state_change_handler('current_sorting_result', handler)

    ###############################################

    def _set_state_value(self, name, val):
        if self._state[name] == val:
            return
        self._state[name] = deepcopy(val)
        self._emit('state-changed-' + name)

    def _get_state_value(self, name):
        return deepcopy(self._state[name])

    def _register_state_change_handler(self, name, handler):
        self._register_signal_handler('state-changed-' + name, handler)

    def _register_signal_handler(self, signal_name, handler):
        if signal_name not in self._signal_handlers:
            self._signal_handlers[signal_name] = []
        self._signal_handlers[signal_name].append(handler)

    def _emit(self, signal_name):
        if signal_name in self._signal_handlers:
            for handler in self._signal_handlers[signal_name]:
                handler()

    def _trigger_any_state_change_handlers(self):
        for handler in self._any_state_change_handlers:
            handler()



import uuid
from spikeforest import mdaio
import io
import base64
import vdomr as vd
import os
import numpy as np
import mtlogging
import time
import traceback

source_path = os.path.dirname(os.path.realpath(__file__))


def _mda32_to_base64(X):
    f = io.BytesIO()
    mdaio.writemda32(X, f)
    return base64.b64encode(f.getvalue()).decode('utf-8')


class TemplateWidget(vd.Component):
    def __init__(self, *, template, size=(200, 200)):
        vd.Component.__init__(self)

        vd.devel.loadBootstrap()
        vd.devel.loadCss(url='https://stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css')
        vd.devel.loadJavascript(path=source_path + '/mda.js')
        vd.devel.loadJavascript(path=source_path + '/canvaswidget.js')
        vd.devel.loadJavascript(path=source_path + '/templatewidget.js')
        vd.devel.loadJavascript(path=source_path + '/../dist/jquery-3.3.1.min.js')

        self._div_id = 'TemplateWidget-' + str(uuid.uuid4())
        self._template = template
        self._template_b64 = _mda32_to_base64(self._template)
        self._y_scale_factor = None

        self._size = size

    def setYScaleFactor(self, scale_factor):
        self._y_scale_factor = scale_factor
        self.refresh()

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        self.refresh()

    def size(self):
        return self._size

    def render(self):
        div = vd.div(id=self._div_id)
        return div

    def postRenderScript(self):
        js = """
        let W=new window.TemplateWidget();
        let X=new window.Mda();
        X.setFromBase64('{template_b64}');
        W.setTemplate(X);
        W.setSize({width},{height});
        W.setYScaleFactor({y_scale_factor});
        $('#{div_id}').empty();
        $('#{div_id}').css({width:'{width}px',height:'{height}px'})
        $('#{div_id}').append(W.element());
        """
        js = js.replace('{template_b64}', self._template_b64)
        js = js.replace('{div_id}', self._div_id)
        js = js.replace('{width}', str(self._size[0]))
        js = js.replace('{height}', str(self._size[1]))
        js = js.replace('{y_scale_factor}', str(self._y_scale_factor or 'null'))
        return js



import spikeextractors as se

import numpy as np
from pathlib import Path

# written for yass v0.5
# James Jun, 2017 Feb 8
# modified from spykingcircussortingextractor.py


class yassSortingExtractor1(se.SortingExtractor):
    def __init__(self, yass_folder):
        se.SortingExtractor.__init__(self)
        yass_folder = Path(yass_folder)
        files = yass_folder.iterdir()
        results = None
        for f in files:
            if 'spike_train.csv' in str(f):
                results = f
                break
        if results is None:
            raise Exception(yass_folder, " is not a yass folder")
        np_results = np.loadtxt(results, delimiter=",")
        spike_times = np_results[:, 0].astype(int)
        spike_clusters = np_results[:, 1].astype(int)

        self._spiketrains = []
        clust_id = np.unique(np_results[:, 1])
        self._unit_ids = list(clust_id)
        for clust in self._unit_ids:
            idx = np.where(spike_clusters == clust)[0]
            self._spiketrains.append(spike_times[idx])

    def get_unit_ids(self):
        return list(self._unit_ids)

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def write_sorting(sorting, save_path):
        # determine the save path
        save_path = Path(save_path)
        if save_path.suffix == '.npy':
            if not str(save_path).endswith('spike_train.npy'):
                raise AttributeError("'save_path' is either a folder or an npy file "
                                     "ending with 'spike_train.npy")
        else:
            if not save_path.is_dir():
                save_path.mkdir()
            save_path = save_path / 'spike_train.npy'

        # write sorting (copyed from kilosortsortingextractor.py)
        spike_times = np.array([])
        spike_templates = np.array([])
        for id in sorting.get_unit_ids():
            st = sorting.get_unit_spike_train(id)
            cl = [id] * len(sorting.get_unit_spike_train(id))
            spike_times = np.concatenate((spike_times, np.array(st)))
            spike_templates = np.concatenate((spike_templates, np.array(cl)))
        sorting_idxs = np.argsort(spike_times)
        spike_times = spike_times[sorting_idxs]
        spike_clusters = spike_templates[sorting_idxs]
        np_results = np.stack((spike_times, spike_clusters), axis=1)
        np.save(save_path, np_results.astype(int))



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
from mountaintools import client as mt
import traceback
import json
from .install_waveclus import install_waveclus


class Waveclus(mlpr.Processor):
    """
    Wave_clus wrapper
      written by J. James Jun, May 21, 2019

    [Optional: Installation instruction in SpikeForest environment]
    1. Run `git clone https://github.com/csn-le/wave_clus.git`
    2. Activate conda environment for SpikeForest
    3. Create `WAVECLUS_PATH_DEV`

    Algorithm website:
    https://github.com/csn-le/wave_clus/wiki
    """

    NAME = 'waveclus'
    VERSION = '0.0.3'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS', 'TEMPDIR']
    ADDITIONAL_FILES = ['*.m', '*.prm']
    CONTAINER = None
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')

    def run(self):
        tmpdir = _get_tmpdir('waveclus')

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            params = read_dataset_params(self.recording_dir)
            # if len(self.channels) > 0:
            #     recording = se.SubRecordingExtractor(
            #         parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)

            all_params = dict()
            for param0 in self.PARAMETERS:
                all_params[param0.name] = getattr(self, param0.name)
            sorting = waveclus_helper(
                recording=recording,
                tmpdir=tmpdir,
                params=params,
                **all_params,
            )
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    print('erased temp file 1')
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            print('erased temp file 2')
            shutil.rmtree(tmpdir)


def waveclus_helper(
        *,
        recording,  # Recording object
        tmpdir,  # Temporary working directory
        params=dict(),
        **kwargs):

    waveclus_path = os.environ.get('WAVECLUS_PATH_DEV', None)
    if waveclus_path:
        print('Using waveclus from WAVECLUS_PATH_DEV directory: {}'.format(waveclus_path))
    else:
        try:
            print('Auto-installing waveclus.')
            waveclus_path = install_waveclus(
                repo='https://github.com/csn-le/wave_clus.git',
                commit='248d15c7eaa2b45b15e4488dfb9b09bfe39f5341'
            )
        except:
            traceback.print_exc()
            raise Exception('Problem installing waveclus. You can set the WAVECLUS_PATH_DEV to force to use a particular path.')
    print('Using waveclus from: {}'.format(waveclus_path))

    # source_dir = os.path.dirname(os.path.realpath(__file__))

    dataset_dir = os.path.join(tmpdir, 'waveclus_dataset')
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.write_recording(
        recording=recording, save_path=dataset_dir, params=params, _preserve_dtype=True)

    samplerate = recording.get_sampling_frequency()

    print('Reading timeseries header...')
    raw_mda = os.path.join(dataset_dir, 'raw.mda')
    HH = mdaio.readmda_header(raw_mda)
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints / samplerate / 60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    # new method
    source_path = os.path.dirname(os.path.realpath(__file__))
    print('Running waveclus in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
        addpath(genpath('{waveclus_path}'), '{source_path}', '{source_path}/mdaio');
        try
            p_waveclus('{tmpdir}', '{dataset_dir}/raw.mda', '{tmpdir}/firings.mda', {samplerate});
        catch
            fprintf('----------------------------------------');
            fprintf(lasterr());
            quit(1);
        end
        quit(0);
    '''
    cmd = cmd.format(waveclus_path=waveclus_path, tmpdir=tmpdir, dataset_dir=dataset_dir, source_path=source_path, samplerate=samplerate)

    matlab_cmd = mlpr.ShellScript(cmd, script_path=tmpdir + '/run_waveclus.m', keep_temp_files=True)
    matlab_cmd.write()

    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        matlab -nosplash -nodisplay -r run_waveclus
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir + '/run_waveclus.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir + '/run_waveclus.sh')
    shell_cmd.start()

    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('waveclus returned a non-zero exit code')

    # parse output
    result_fname = tmpdir + '/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.set_times_labels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)


def _run_command_and_print_output(command):
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def read_dataset_params(dsdir):
    # ca = _load_required_modules()
    fname1 = dsdir + '/params.json'
    fname2 = mt.realizeFile(path=fname1)
    if not fname2:
        raise Exception('Unable to find file: ' + fname1)
    if not os.path.exists(fname2):
        raise Exception('Dataset parameter file does not exist: ' + fname2)
    with open(fname2) as f:
        return json.load(f)


# To be shared across sorters (2019.05.05)
def _get_tmpdir(sorter_name):
    code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
    tmpdir0 = os.environ.get('TEMPDIR', '/tmp')
    tmpdir = os.path.join(tmpdir0, '{}-tmp-{}'.format(sorter_name, code))
    # reset the output folder
    if os.path.exists(tmpdir):
        shutil.rmtree(str(tmpdir))
    else:
        os.makedirs(tmpdir)
    return tmpdir



import vdomr as vd
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging
import numpy as np
import spikeforest_analysis as sa
from spikeforest import EfficientAccessRecordingExtractor
import json


class UnitTableView(vd.Component):
    def __init__(self, context):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._unit_table_widget = None

        self._connection_to_init, connection_to_parent = multiprocessing.Pipe()
        self._init_process = multiprocessing.Process(target=_initialize, args=(context, connection_to_parent))
        self._init_process.start()

        self._init_log_text = ''
        vd.set_timeout(self._check_init, 0.5)

    def _on_init_completed(self, result):
        self._unit_table_widget = UnitTableWidget(context=self._context, units_info=result)
        self._unit_table_widget.setSize(self._size)
        self.refresh()

    def setSize(self, size):
        self._size = size
        if self._unit_table_widget:
            self._unit_table_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Unit table'

    def render(self):
        if self._unit_table_widget:
            return vd.div(
                self._unit_table_widget
            )
        else:
            return vd.div(
                vd.h3('Initializing...'),
                vd.pre(self._init_log_text)
            )

    def _check_init(self):
        if not self._unit_table_widget:
            if self._connection_to_init.poll():
                msg = self._connection_to_init.recv()
                if msg['name'] == 'log':
                    self._init_log_text = self._init_log_text + msg['text']
                    self.refresh()
                elif msg['name'] == 'result':
                    self._on_init_completed(msg['result'])
                    return
            vd.set_timeout(self._check_init, 1)


class UnitTableWidget(vd.Component):
    def __init__(self, *, context, units_info):
        vd.Component.__init__(self)
        self._size = (100, 100)
        self._units_info = units_info

    def setSize(self, size):
        self._size = size
        self.refresh()

    def render(self):
        return vd.pre(json.dumps(self._units_info, indent=2))

# Initialization in a worker thread
mtlogging.log(root=True)


def _initialize(context, connection_to_parent):
    with StdoutSender(connection=connection_to_parent):
        print('***** Preparing efficient access recording extractor...')
        _ = EfficientAccessRecordingExtractor(recording=context.recordingExtractor())
        print('***** computing units info...')
        info0 = sa.compute_units_info(recording=context.recordingExtractor(), sorting=context.trueSortingExtractor())
        print('*****')
    connection_to_parent.send(dict(
        name='result',
        result=info0
    ))



import mlprocessors as mlpr

import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from .yasssortingextractor1 import yassSortingExtractor1
from .tools import saveProbeFile
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import spikeextractors as se

# yass uses negative polarity by default


class YASS1(mlpr.Processor):
    NAME = 'YASS1'
    VERSION = '0.0.1'
    # used by container to pass the env variables
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    ADDITIONAL_FILES = ['*.yaml']

    # The following container uses python 2
    # CONTAINER = 'sha1://087767605e10761331699dda29519444bbd823f4/02-12-2019/yass.simg'

    # this one uses python 3
    CONTAINER = 'sha1://85f443e414ac454f2057553bf6436131d0eb6245/yass1.simg'
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')
    # paramfile_out = mlpr.Output('YASS yaml config file')

    detect_sign = mlpr.IntegerParameter(optional=True, default=-1,
                                        description='-1, 1, or 0')
    adjacency_radius = mlpr.FloatParameter(
        optional=True, default=70, description='Channel neighborhood adjacency radius corresponding to geom file')
    template_width_ms = mlpr.FloatParameter(
        optional=True, default=1, description='Spike width in milliseconds')
    filter = mlpr.BoolParameter(optional=True, default=True)

    def run(self):
        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp') + '/yass-tmp-' + code

        # num_workers = os.environ.get('NUM_WORKERS', 1)
        # print('num_workers: {}'.format(num_workers))
        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting, _ = yass_helper(
                recording=recording,
                output_folder=tmpdir,
                probe_file=None,
                file_name=None,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                template_width_ms=self.template_width_ms,
                filter=self.filter)
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
            # shutil.copyfile(yaml_file, self.paramfile_out)
        except:
            if os.path.exists(tmpdir):
                # shutil.rmtree(tmpdir)
                print('not deleted tmpdir1')
            raise
        if not getattr(self, '_keep_temp_files', False):
            # shutil.rmtree(tmpdir)
            print('not deleted tmpdir2')


def yass_helper(
        recording,
        output_folder=None,  # Temporary working directory
        probe_file=None,
        file_name=None,
        detect_sign=-1,  # -1 - 1 - 0
        template_width_ms=1,  # yass parameter
        filter=True,
        adjacency_radius=100):
    source_dir = os.path.dirname(os.path.realpath(__file__))

    # make output dir
    if output_folder is None:
        output_folder = 'yass'
    else:
        output_folder = join(output_folder, 'yass')
    output_folder = os.path.abspath(output_folder)
    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # save prb file:
    if probe_file is None:
        probe_file = join_abspath_(output_folder, 'probe.npy')
    saveProbeFile(recording, probe_file, format='yass')

    # save binary file
    if file_name is None:
        file_name = 'raw.bin'
    bin_file = join_abspath_(output_folder, file_name)
    # print('bin_file:{}'.format(bin_file))
    write_recording_(recording=recording, save_path=bin_file, fReversePolarity=(detect_sign > 0), dtype=np.float32, scale_factor=1)
    # print('bin_file exists? {}'.format(os.path.exists(bin_file)))

    # set up yass config file
    print(source_dir)
    with open(join(source_dir, 'config_default.yaml'), 'r') as f:
        yass_config = f.read()

    # get the order
    # root_folder, recordings, geometry, dtype, sampling_rate, n_channels, spatial_radius, spike_size_ms, filter
    n_channels = recording.get_num_channels()
    sampling_rate = recording.get_sampling_frequency()

    # print('sampling_rate={}'.format(sampling_rate))

    yaml_file = join(output_folder, file_name + '.yaml')
    yass_config = yass_config.format(
        output_folder, bin_file, probe_file, 'single', int(sampling_rate), n_channels, adjacency_radius, template_width_ms, filter)
    with open(yaml_file, 'w') as f:
        f.write(yass_config)

    with open(yaml_file) as ff:
        print('YASS CONFIG:')
        print(ff.read())

    print('Running yass...')
    t_start_proc = time.time()

    # yass_path = '/usr/local/bin'
    # num_cores_str = ''
    # cmd = 'python2 {}/yass {} {} '.format(
    #    yass_path, join(output_folder, file_name+'.yaml'), num_cores_str)
    cmd = 'yass {}'.format(join(output_folder, file_name + '.yaml'))

    retcode = run_command_and_print_output(cmd)
    if retcode != 0:
        raise Exception('yass returned a non-zero exit code')

    # retcode = run_command_and_print_output(cmd_merge)
    # if retcode != 0:
    #    raise Exception('yass merging returned a non-zero exit code')
    processing_time = time.time() - t_start_proc
    print('Elapsed time: ', processing_time)
    sorting = yassSortingExtractor1(output_folder)

    return sorting, yaml_file


def run_command_and_print_output(command):
    print('RUNNING: ' + command)
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def join_abspath_(path1, path2):
    path_abs = os.path.abspath(os.path.join(path1, path2))
    return path_abs


def write_recording_(recording, save_path, dtype=None, transpose=False, fReversePolarity=False, scale_factor=1):
    # save_path = Path(save_path)
    print('write_recording2: {}'.format(str(save_path)))

    if dtype is None:
        dtype = np.float32
    np_Wav = np.array(recording.get_traces(), dtype=dtype)
    if transpose:
        np_Wav = np.transpose(np_Wav)
    if fReversePolarity:
        np_Wav = np_Wav * -1
    np_Wav = np_Wav * scale_factor
    with open(save_path, 'wb') as f:
        np.ravel(np_Wav, order='F').tofile(f)



import vdomr as vd
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging
import numpy as np
from spikeforest import EfficientAccessRecordingExtractor
import json
import mlprocessors as mlpr
import traceback
from mountaintools import client as mt
import pickle


class UnitDetailView(vd.Component):
    def __init__(self, context):
        vd.Component.__init__(self)
        self._context = context
        self._unit_id = context.currentUnitId()
        self._size = (100, 100)
        self._unit_detail_widget = None

        if self._unit_id is None:
            return

        self._connection_to_init, connection_to_parent = multiprocessing.Pipe()
        self._init_process = multiprocessing.Process(target=_initialize, args=(context, self._unit_id, connection_to_parent))
        self._init_process.start()

        self._init_log_text = ''
        vd.set_timeout(self._check_init, 0.5)

    def _on_init_completed(self, result):
        self._unit_detail_widget = UnitDetailWidget(context=self._context, unit_id=self._unit_id, init_result=result)
        self._unit_detail_widget.setSize(self._size)
        self.refresh()

    def setSize(self, size):
        self._size = size
        if self._unit_detail_widget:
            self._unit_detail_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Unit {}'.format(self._unit_id)

    def render(self):
        if self._unit_id is None:
            return vd.div(vd.h3('No current unit selected.'))
        if self._unit_detail_widget:
            return vd.div(
                self._unit_detail_widget
            )
        else:
            return vd.div(
                vd.h3('Initializing......'),
                vd.pre(self._init_log_text),
                style=dict(overflow='auto')
            )

    def _check_init(self):
        if not self._unit_detail_widget:
            if self._connection_to_init.poll():
                msg = self._connection_to_init.recv()
                if msg['name'] == 'log':
                    self._init_log_text = self._init_log_text + msg['text']
                    self.refresh()
                elif msg['name'] == 'result':
                    self._on_init_completed(msg['result'])
                    return
            vd.set_timeout(self._check_init, 1)


class UnitDetailWidget(vd.Component):
    def __init__(self, *, context, unit_id, init_result):
        vd.Component.__init__(self)
        self._context = context
        self._unit_id = unit_id
        self._init_result = init_result
        self._size = (100, 100)

    def setSize(self, size):
        self._size = size
        self.refresh()

    def render(self):
        result0 = self._init_result
        snippets = result0['snippets']
        template = result0['template']
        peak_chan_index = np.argmax(np.max(np.abs(template), axis=1), axis=0)
        peak_chan = self._context.recordingExtractor().get_channel_ids()[peak_chan_index]
        return vd.div(
            vd.pre(
                'Unit {} has {} events. Peak is on channel {}.'.format(self._unit_id, result0['num_events'], peak_chan)
            ),
            SnippetsPlot(snippets=snippets, channel_index=peak_chan_index, size=(self._size[0], self._size[1] - 50))
        )


class SnippetsPlot(vd.Component):
    def __init__(self, *, snippets, channel_index, size):
        vd.Component.__init__(self)
        data = []
        for snippet in snippets:
            # M = snippet.shape[0]
            T = snippet.shape[1]
            data.append(dict(
                x=np.arange(0, T),
                y=snippet[channel_index, :]
            ))
        self._plot = vd.components.PlotlyPlot(
            data=data,
            layout=dict(
                showlegend=False
            ),
            config=dict(),
            size=size
        )

    def render(self):
        return vd.div(
            self._plot
        )


# Initialization in a worker thread
mtlogging.log(root=True)


def _initialize(context, unit_id, connection_to_parent):
    with StdoutSender(connection=connection_to_parent):
        try:
            print('***** Preparing efficient access recording extractor...')
            earx = EfficientAccessRecordingExtractor(recording=context.recordingExtractor())
            print('***** Computing unit detail...')
            path0 = mt.realizeFile(path=ComputeUnitDetail.execute(recording=earx, sorting=context.trueSortingExtractor(), unit_id=unit_id, output=True).outputs['output'])
            with open(path0, 'rb') as f:
                result0 = pickle.load(f)
            print('*****')
        except:
            traceback.print_exc()
            raise
    connection_to_parent.send(dict(
        name='result',
        result=result0
    ))


class ComputeUnitDetail(mlpr.Processor):
    NAME = 'ComputeUnitsDetail'
    VERSION = '0.1.5'
    recording = mlpr.Input()
    sorting = mlpr.Input()
    unit_id = mlpr.IntegerParameter()
    output = mlpr.Output()

    def run(self):
        event_times = self.sorting.get_unit_spike_train(unit_id=self.unit_id)  # pylint: disable=no-member
        snippets = self.recording.get_snippets(reference_frames=event_times, snippet_len=100)  # pylint: disable=no-member
        template = np.median(np.stack(snippets), axis=0)
        result0 = dict(
            unit_id=self.unit_id,
            num_events=len(event_times),
            event_times=event_times,
            snippets=snippets,
            template=template
        )
        with open(self.output, 'wb') as f:
            pickle.dump(result0, f)



import vdomr as vd
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging
import numpy as np


class TestPlotlyView(vd.Component):
    def __init__(self, context):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._test_plotly_widget = None

        self._connection_to_init, connection_to_parent = multiprocessing.Pipe()
        self._init_process = multiprocessing.Process(target=_initialize, args=(context, connection_to_parent))
        self._init_process.start()

        self._init_log_text = ''
        vd.set_timeout(self._check_init, 0.5)

    def _on_init_completed(self, init):
        self._test_plotly_widget = TestPlotlyWidget()
        self._test_plotly_widget.setSize(self._size)
        self.refresh()

    def setSize(self, size):
        self._size = size
        if self._test_plotly_widget:
            self._test_plotly_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Test plotly'

    def render(self):
        if self._test_plotly_widget:
            return vd.div(
                self._test_plotly_widget
            )
        else:
            return vd.div(
                vd.h3('Initializing...'),
                vd.pre(self._init_log_text)
            )

    def _check_init(self):
        if not self._test_plotly_widget:
            if self._connection_to_init.poll():
                msg = self._connection_to_init.recv()
                if msg['name'] == 'log':
                    self._init_log_text = self._init_log_text + msg['text']
                    self.refresh()
                elif msg['name'] == 'result':
                    self._on_init_completed(msg['result'])
                    return
            vd.set_timeout(self._check_init, 1)


class TestPlotlyWidget(vd.Component):
    def __init__(self):
        vd.Component.__init__(self)
        self._size = (100, 100)
        self._plot = None
        self._update_plot()

    def setSize(self, size):
        self._size = size
        self._update_plot()

    def _update_plot(self):
        xx = np.linspace(0, 1, 10)
        yy = np.cos((10 * xx)**2)
        self._plot = vd.components.PlotlyPlot(
            data=dict(x=xx, y=yy),
            layout=dict(margin=dict(t=5)),
            config=dict(),
            size=self._size
        )
        self.refresh()

    def render(self):
        if not self._plot:
            return vd.div('no plot.')
        return self._plot

# Initialization in a worker thread
mtlogging.log(root=True)


def _initialize(context, connection_to_parent):
    with StdoutSender(connection=connection_to_parent):
        pass
    connection_to_parent.send(dict(
        name='result',
        result=dict()
    ))



from abc import ABC, abstractmethod
import spikeextractors as se
import numpy as np


class LazyFilterRecording(se.RecordingExtractor):
    def __init__(self, *, recording, chunk_size=10000):
        se.RecordingExtractor.__init__(self)
        self._recording = recording
        self._chunk_size = chunk_size
        self._filtered_chunk_cache = FilteredChunkCache()
        self.copy_channel_properties(recording)

    def get_channel_ids(self):
        return self._recording.get_channel_ids()

    def get_num_frames(self):
        return self._recording.get_num_frames()

    def get_sampling_frequency(self):
        return self._recording.get_sampling_frequency()

    def get_traces(self, start_frame=None, end_frame=None, channel_ids=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        all_channel_ids = self.get_channel_ids()
        if channel_ids is None:
            channel_ids = all_channel_ids
        channel_indices = [all_channel_ids.index(ch) for ch in channel_ids]
        ich1 = int(start_frame / self._chunk_size)
        ich2 = int((end_frame - 1) / self._chunk_size)
        filtered_chunk_list = []
        for ich in range(ich1, ich2 + 1):
            filtered_chunk0 = self._get_filtered_chunk(ich)
            if ich == ich1:
                start0 = start_frame - ich * self._chunk_size
            else:
                start0 = 0
            if ich == ich2:
                end0 = end_frame - ich * self._chunk_size
            else:
                end0 = self._chunk_size
            filtered_chunk_list.append(filtered_chunk0[channel_indices, start0:end0])
        return np.concatenate(filtered_chunk_list, axis=1)

    @abstractmethod
    def filterChunk(self, *, start_frame, end_frame):
        raise NotImplementedError('filterChunk not implemented')

    def _get_filtered_chunk(self, ind):
        code = str(ind)
        chunk0 = self._filtered_chunk_cache.get(code)
        if chunk0 is not None:
            return chunk0
        else:
            start0 = ind * self._chunk_size
            end0 = (ind + 1) * self._chunk_size
            chunk1 = self.filterChunk(start_frame=start0, end_frame=end0)
            self._filtered_chunk_cache.add(code, chunk1)
            return chunk1


class FilteredChunkCache():
    def __init__(self):
        self._chunks_by_code = dict()
        self._codes = []
        self._total_size = 0
        self._max_size = 1024 * 1024 * 100

    def add(self, code, chunk):
        self._chunks_by_code[code] = chunk
        self._codes.append(code)
        self._total_size = self._total_size + chunk.size
        if self._total_size > self._max_size:
            ii = 0
            while (ii < len(self._codes)) and (self._total_size > self._max_size / 2):
                self._total_size = self._total_size - self._chunks_by_code[self._codes[ii]].size
                del self._chunks_by_code[self._codes[ii]]
                ii = ii + 1
            self._codes = self._codes[ii:]

    def get(self, code):
        if code in self._chunks_by_code:
            return self._chunks_by_code[code]
        else:
            return None



from mountaintools import client as mt
from PIL import Image
import json
import pandas as pd
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import spikeextractors as si


def kb_read_text_file(fname):
    return mt.loadText(path=fname)


def kb_read_json_file(fname):
    return mt.loadObject(path=fname)


class SFSortingResult():
    def __init__(self, obj, recording):
        self._obj = obj
        self._recording = recording

    def getObject(self):
        return self._obj

    def recording(self):
        return self._recording

    def sorterName(self):
        return self._obj['sorter']['name']

    def plotNames(self):
        plots = self._obj['summary'].get('plots', dict())
        return list(plots.keys())

    def sorting(self):
        return SFMdaSortingExtractor(firings_file=self._obj['firings'])

    def plot(self, name, format='image'):
        plots = self._obj['summary'].get('plots', dict())
        url = plots[name]
        if format == 'url':
            return url
        else:
            path = mt.realizeFile(url)
            if format == 'image':
                return Image.open(path)
            elif format == 'path':
                return path
            else:
                raise Exception('Invalid format: ' + format)

    def comparisonWithTruth(self, *, format='dataframe'):
        A = self._obj['comparison_with_truth']
        if not A:
            return None
        if format == 'html':
            return kb_read_text_file(A['html'])
        else:
            B = kb_read_json_file(A['json'])
            if format == 'json':
                return B
            elif format == 'dataframe':
                return pd.DataFrame(B).transpose()
            else:
                raise Exception('Invalid format: ' + format)


class SFRecording():
    def __init__(self, obj, study):
        self._obj = obj
        self._sorting_result_names = []
        self._sorting_results_by_name = dict()
        self._summary_result = None
        if 'summary' in obj:
            self._summary_result = obj['summary']
        self._study = study

    def getObject(self):
        return self._obj

    def getSummaryObject(self):
        return self._summary_result

    def study(self):
        return self._study

    def name(self):
        return self._obj.get('recording_name', self._obj.get('name'))

    def description(self):
        return self._obj['description']

    def directory(self):
        return self._obj['directory']

    def recordingFileIsLocal(self):
        fname = self.directory() + '/raw.mda'
        fname2 = mt.findFile(fname, local_only=True)
        if fname2 and (not _is_url(fname2)):
            return True
        return False

    def realizeRecordingFile(self):
        fname = self.directory() + '/raw.mda'
        return mt.realizeFile(fname)

    def firingsTrueFileIsLocal(self):
        fname = self.directory() + '/firings_true.mda'
        fname2 = mt.findFile(fname, local_only=True)
        if fname2 and (not _is_url(fname2)):
            return True
        return False

    def realizeFiringsTrueFile(self):
        fname = self.directory() + '/firings_true.mda'
        return mt.realizeFile(fname)

    def recordingExtractor(self, download=False):
        X = SFMdaRecordingExtractor(dataset_directory=self.directory(), download=download)
        if 'channels' in self._obj:
            if self._obj['channels']:
                X = si.SubRecordingExtractor(parent_recording=X, channel_ids=self._obj['channels'])
        return X

    def sortingTrue(self):
        return SFMdaSortingExtractor(firings_file=self.directory() + '/firings_true.mda')

    def plotNames(self):
        if not self._summary_result:
            return []
        plots = self._summary_result.get('plots', dict())
        return list(plots.keys())

    def plot(self, name, format='image'):
        plots = self._summary_result.get('plots', dict())
        url = plots[name]
        if format == 'url':
            return url
        else:
            path = mt.realizeFile(url)
            if format == 'image':
                return Image.open(path)
            elif format == 'path':
                return path
            else:
                raise Exception('Invalid format: ' + format)

    def trueUnitsInfo(self, format='dataframe'):
        if not self._summary_result:
            return None
        B = kb_read_json_file(self._summary_result['true_units_info'])
        if format == 'json':
            return B
        elif format == 'dataframe':
            return pd.DataFrame(B)
        else:
            raise Exception('Invalid format: ' + format)

    def setSummaryResult(self, obj):
        self._summary_result = obj

    def addSortingResult(self, obj):
        sorter_name = obj['sorter']['name']
        if sorter_name in self._sorting_results_by_name:
            print('Sorting result already in recording: {}'.format(sorter_name))
        else:
            R = SFSortingResult(obj, self)
            self._sorting_result_names.append(sorter_name)
            self._sorting_results_by_name[sorter_name] = R

    def sortingResultNames(self):
        return self._sorting_result_names

    def sortingResult(self, name):
        return self._sorting_results_by_name.get(name, None)


class SFStudy():
    def __init__(self, obj):
        self._obj = obj
        self._recordings_by_name = dict()
        self._recording_names = []

    def getObject(self):
        return self._obj

    def name(self):
        return self._obj.get('recording_name', self._obj.get('name'))

    def description(self):
        return self._obj['description']

    def addRecording(self, obj):
        name = obj.get('recording_name', obj.get('name'))
        if name in self._recordings_by_name:
            print('Recording already in study: ' + name)
        else:
            self._recording_names.append(name)
            D = SFRecording(obj, self)
            self._recordings_by_name[name] = D

    def recordingNames(self):
        return self._recording_names

    def recording(self, name):
        return self._recordings_by_name.get(name, None)


class SFData():
    def __init__(self):
        self._studies_by_name = dict()
        self._study_names = []

    def loadStudy(self, study):
        name = study['name']
        if name in self._studies_by_name:
            print('Study already loaded: ' + name)
        else:
            self._study_names.append(study['name'])
            S = SFStudy(study)
            self._studies_by_name[name] = S

    def loadStudies(self, studies):
        for study in studies:
            self.loadStudy(study)

    def loadRecording(self, recording):
        study = recording.get('study_name', recording.get('study_name', recording.get('study')))
        self._studies_by_name[study].addRecording(recording)

    def loadRecordings2(self, recordings):
        for recording in recordings:
            self.loadRecording(recording)

    def loadRecordings(self, *, key=None, verbose=False):
        # old
        if key is None:
            key = dict(name='spikeforest_studies_processed')
        obj = mt.loadObject(key=key)
        studies = obj['studies']
        for study in studies:
            name = study['name']
            if name in self._studies_by_name:
                print('Study already loaded: ' + name)
            else:
                self._study_names.append(study['name'])
                S = SFStudy(study)
                self._studies_by_name[name] = S
        recordings = obj['recordings']
        print('recordings ===================================================================')
        print(recordings)
        for ds in recordings:
            study = ds.get('study_name', ds.get('study'))
            self._studies_by_name[study].addRecording(ds)
        if verbose:
            print('Loaded {} recordings'.format(len(recordings)))

    def loadSortingResults(self, sorting_results):
        for result in sorting_results:
            self.loadSortingResult(result)

    def loadSortingResult(self, X):
        study_name = X['recording'].get('study_name', X['recording'].get('study'))
        recording_name = X['recording'].get('recording_name', X['recording'].get('name'))
        # sorter_name=X['sorter']['name']
        S = self.study(study_name)
        if S:
            D = S.recording(recording_name)
            if D:
                D.addSortingResult(X)
            else:
                print('Warning: recording not found: ' + recording_name)
        else:
            print('Warning: study not found: ' + study_name)

    def studyNames(self):
        return self._study_names

    def study(self, name):
        return self._studies_by_name.get(name, None)


def _is_url(path):
    return (path.startswith('http://') or path.startswith('https://'))



import numpy as np


def exp_growth(amp1, amp2, dur1, dur2):
    t = np.arange(0, dur1)
    Y = np.exp(t / dur2)
    # Want Y[0]=amp1
    # Want Y[-1]=amp2
    Y = Y / (Y[-1] - Y[0]) * (amp2 - amp1)
    Y = Y - Y[0] + amp1
    return Y


def exp_decay(amp1, amp2, dur1, dur2):
    Y = exp_growth(amp2, amp1, dur1, dur2)
    Y = np.flipud(Y)  # used to be flip, but that was not supported by older versions of numpy
    return Y


def smooth_it(Y, t):
    Z = np.zeros(Y.size)
    for j in range(-t, t + 1):
        Z = Z + np.roll(Y, j)
    return Z


def synthesize_single_waveform(*, N=800, durations=[200, 10, 30, 200], amps=[0.5, 10, -1, 0]):
    durations = np.array(durations).ravel()
    if (np.sum(durations) >= N - 2):
        durations[-1] = N - 2 - np.sum(durations[0:durations.size - 1])

    amps = np.array(amps).ravel()

    timepoints = np.round(np.hstack((0, np.cumsum(durations) - 1))).astype('int')

    t = np.r_[0:np.sum(durations) + 1]

    Y = np.zeros(len(t))
    Y[timepoints[0]:timepoints[1] + 1] = exp_growth(0, amps[0], timepoints[1] + 1 - timepoints[0], durations[0] / 4)
    Y[timepoints[1]:timepoints[2] + 1] = exp_growth(amps[0], amps[1], timepoints[2] + 1 - timepoints[1], durations[1])
    Y[timepoints[2]:timepoints[3] + 1] = exp_decay(amps[1], amps[2], timepoints[3] + 1 - timepoints[2],
                                                   durations[2] / 4)
    Y[timepoints[3]:timepoints[4] + 1] = exp_decay(amps[2], amps[3], timepoints[4] + 1 - timepoints[3],
                                                   durations[3] / 5)
    Y = smooth_it(Y, 3)
    Y = Y - np.linspace(Y[0], Y[-1], len(t))
    Y = np.hstack((Y, np.zeros(N - len(t))))
    Nmid = int(np.floor(N / 2))
    peakind = np.argmax(np.abs(Y))
    Y = np.roll(Y, Nmid - peakind)

    return Y


# Y=smooth_it(Y,3);
# Y=Y-linspace(Y(1),Y(end),length(Y));
#
# Y=[Y,zeros(1,N-length(Y))];
#
# Nmid=floor(N/2);
# [~,peakind]=max(abs(Y));
# Y=circshift(Y,[0,Nmid-peakind]);
#
# end
#
# function test_synth_waveform
# Y=synthesize_single_waveform(800);
# figure; plot(Y);
# end
#
# function Y=exp_growth(amp1,amp2,dur1,dur2)
# t=1:dur1;
# Y=exp(t/dur2);
# % Want Y(1)=amp1
# % Want Y(end)=amp2
# Y=Y/(Y(end)-Y(1))*(amp2-amp1);
# Y=Y-Y(1)+amp1;
# end
#
# function Y=exp_decay(amp1,amp2,dur1,dur2)
# Y=exp_growth(amp2,amp1,dur1,dur2);
# Y=Y(end:-1:1);
# end
#
# function Z=smooth_it(Y,t)
# Z=Y;
# Z(1+t:end-t)=0;
# for j=-t:t
#    Z(1+t:end-t)=Z(1+t:end-t)+Y(1+t+j:end-t+j)/(2*t+1);
# end;
# end

if __name__ == '__main__':
    Y = synthesize_single_waveform()
    import matplotlib.pyplot as plt

    plt.plot(Y)



from matplotlib import pyplot as plt
import numpy as np

class UnitWaveformsWidget:
    def __init__(self, *, recording, sorting, channels=None, unit_ids=None, width=14, height=7, snippet_len=100,
                 title='', max_num_spikes_per_unit=50):
        self._IX = recording
        self._OX = sorting
        self._channels = channels
        self._unit_ids = unit_ids
        self._width = width
        self._height = height
        self._figure = None
        self._snippet_len = snippet_len
        self._title = title
        self._max_num_spikes_per_unit = max_num_spikes_per_unit

    def plot(self):
        self._do_plot()

    def figure(self):
        return self._figure

    def _do_plot(self):
        units = self._unit_ids
        channels = self._channels
        if units is None:
            units = self._OX.get_unit_ids()
        channel_ids = self._IX.get_channel_ids()
        M = len(channel_ids)
        channel_locations = np.zeros((M, 2))
        for ii, ch in enumerate(channel_ids):
            loc = self._IX.get_channel_property(ch, 'location')
            channel_locations[ii, :] = loc[-2:]
        if channels is None:
            channels = channel_ids
        list = []
        for unit in units:
            st = self._OX.get_unit_spike_train(unit_id=unit)
            if st is not None:
                spikes = self._get_random_spike_waveforms(unit=unit, max_num=self._max_num_spikes_per_unit,
                                                          channels=channels)
                item = dict(
                    representative_waveforms=spikes,
                    title='Unit {}'.format(int(unit))
                )
                list.append(item)
            else:
                print(unit, ' spike train is None')
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            # self._plot_spike_shapes_multi(list,channel_locations=channel_locations[np.array(channels),:])
            self._plot_spike_shapes_multi(list, channel_locations=None)

    def _get_random_spike_waveforms(self, *, unit, max_num, channels):
        st = self._OX.get_unit_spike_train(unit_id=unit)
        num_events = len(st)
        if num_events > max_num:
            event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
        else:
            event_indices = range(num_events)

        spikes = self._IX.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=self._snippet_len,
                                      channel_ids=channels)
        if len(spikes)>0:
            spikes = np.dstack(tuple(spikes))
        else:
            spikes = np.zeros((self._IX.get_num_channels(), self._snippet_len, 0))
        return spikes

    def _plot_spike_shapes(self, *, representative_waveforms=None, average_waveform=None, channel_locations=None,
                           ylim=None, max_representatives=None, color='blue', title=''):
        if average_waveform is None:
            if representative_waveforms is None:
                raise Exception('You must provide either average_waveform, representative waveforms, or both')
            average_waveform = np.mean(representative_waveforms, axis=2)
        M = average_waveform.shape[0]  # number of channels
        T = average_waveform.shape[1]  # number of timepoints
        if ylim is None:
            ylim = [average_waveform.min(), average_waveform.max()]
        yrange = ylim[1] - ylim[0]
        if channel_locations is None:
            channel_locations = np.zeros((M, 2))
            for m in range(M):
                channel_locations[m, :] = [0, -m]

        spacing = 1 / 0.8  # TODO: auto-determine this from the channel_locations

        xvals = np.linspace(-yrange / 2, yrange / 2, T)
        if representative_waveforms is not None:
            if max_representatives is not None:
                W0 = representative_waveforms
                if W0.shape[2] > max_representatives:
                    indices = np.random.choice(range(W0.shape[2]), size=max_representatives, replace=False)
                    representative_waveforms = W0[:, :, indices]
            L = representative_waveforms.shape[2]
            XX = np.zeros((T, M, L))
            YY = np.zeros((T, M, L))
            for m in range(M):
                loc = channel_locations[m, -2:] * yrange * spacing
                for j in range(L):
                    XX[:, m, j] = loc[0] + xvals
                    YY[:, m, j] = loc[1] + representative_waveforms[m, :, j] - representative_waveforms[m, 0, j]
            XX = XX.reshape(T, M * L)
            YY = YY.reshape(T, M * L)
            plt.plot(XX, YY, color=(0.5, 0.5, 0.5), alpha=0.4)

            XX = np.zeros((T, M))
            YY = np.zeros((T, M))
            for m in range(M):
                loc = channel_locations[m, -2:] * yrange * spacing
                XX[:, m] = loc[0] + xvals
                YY[:, m] = loc[1] + average_waveform[m, :] - average_waveform[m, 0]
            plt.plot(XX, YY, color)

        plt.gca().get_xaxis().set_ticks([])
        plt.gca().get_yaxis().set_ticks([])
        if title:
            plt.title(title, color='gray')

    def _get_ylim_for_item(self, average_waveform=None, representative_waveforms=None):
        if average_waveform is None:
            if representative_waveforms is None:
                raise Exception('You must provide either average_waveform, representative waveforms, or both')
            average_waveform = np.mean(representative_waveforms, axis=2)
        return [average_waveform.min(), average_waveform.max()]

    def _determine_global_ylim(self, list):
        ret = [0, 0]
        for item in list:
            ylim0 = self._get_ylim_for_item(
                average_waveform=item.get('average_waveform', None),
                representative_waveforms=item.get('representative_waveforms', None)
            )
            ret[0] = np.minimum(ylim0[0], ret[0])
            ret[1] = np.maximum(ylim0[1], ret[1])
        return ret

    def _plot_spike_shapes_multi(self, list, *, ncols=5, **kwargs):
        if 'ylim' in kwargs:
            ylim = kwargs['ylim']
        else:
            ylim = self._determine_global_ylim(list)
        nrows = np.ceil(len(list) / ncols)
        self._figure = plt.figure(figsize=(3 * ncols + 0.1, 3 * nrows + 0.1))
        for i, item in enumerate(list):
            plt.subplot(nrows, ncols, i + 1)
            self._plot_spike_shapes(**item, **kwargs, ylim=ylim)



import vdomr as vd
import time
import sys
import mtlogging
from mountaintools import client as mt
from .tablewidget import TableWidget
import json


class SortingResultDetailView(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._sr_context = context
        self._size = (100, 100)
        self._widget = SortingResultDetailWidget(
            context=self._sr_context
        )
        self._widget.setSize(self._size)

    def setSize(self, size):
        self._size = size
        if self._widget:
            self._widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Sorting result detail'

    def render(self):
        return self._widget


class SortingResultDetailWidget(vd.Component):
    def __init__(self, *, context):
        vd.Component.__init__(self)
        self._sr_context = context
        self._size = (100, 100)

    def setSize(self, size):
        self._size = size

    def render(self):
        sr = self._sr_context.sortingResultObject()
        return vd.components.ScrollArea(vd.pre(json.dumps(sr, indent=4)), height=self._size[1])



from pathlib import Path

import mlprocessors as mlpr
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import os
import time
import random
import string
import shutil
import sys
import shlex
import json
from mountaintools import client as mt
from subprocess import Popen, PIPE, CalledProcessError, call


class Klusta(mlpr.Processor):
    """

    Installation instruction
        >>> pip install Cython h5py tqdm
        >>> pip install click klusta klustakwik2

    More information on klusta at:
      * https://github.com/kwikteam/phy"
      * https://github.com/kwikteam/klusta

    """

    NAME = 'Klusta'
    VERSION = '0.1.1'  # wrapper VERSION
    ADDITIONAL_FILES = ['*.prm']
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS', 'TEMPDIR']
    CONTAINER = 'sha1://6d76f22e3b4eff52b430ef4649a8802f7da9e0ec/2019-05-13/klusta.simg'
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')
    channels = mlpr.IntegerListParameter(optional=True, default=[],
                                         description='List of channels to use.')
    adjacency_radius = mlpr.FloatParameter(optional=True, default=None, description='')
    detect_sign = mlpr.FloatParameter(optional=True, default=-1, description='')
    threshold_strong_std_factor = mlpr.FloatParameter(optional=True, default=5, description='')
    threshold_weak_std_factor = mlpr.FloatParameter(optional=True, default=2, description='')
    n_features_per_channel = mlpr.IntegerParameter(optional=True, default=3, description='')
    num_starting_clusters = mlpr.IntegerParameter(optional=True, default=3, description='')
    extract_s_before = mlpr.IntegerParameter(optional=True, default=16, description='')
    extract_s_after = mlpr.IntegerParameter(optional=True, default=32, description='')

    def run(self):
        try:
            import klusta
            import klustakwik2
            HAVE_KLUSTA = True
        except ImportError:
            HAVE_KLUSTA = False

        if not HAVE_KLUSTA:
            raise Exception('Klusta kwik is not installed.')

        # alias to params
        p = {
            'probe_file': None,
            'pca_n_waveforms_max': 10000,
        }
        for param0 in self.PARAMETERS:  # pylint: disable=no-member
            p[param0.name] = getattr(self, param0.name)
        source_dir = Path(__file__).parent

        tmpdir = Path(_get_tmpdir('klusta'))

        # source file
        recording = SFMdaRecordingExtractor(self.recording_dir)
        if len(self.channels) > 0:
            recording = se.SubRecordingExtractor(
                parent_recording=recording, channel_ids=self.channels)

        # if isinstance(recording, se.BinDatRecordingExtractor) and recording._frame_first and\
        #                 recording._timeseries.offset == 0:
        #     # no need to copy
        #     raw_filename = str(recording._datfile)
        #     raw_filename = raw_filename.replace('.dat', '')
        #     dtype = recording._timeseries.dtype.str
        #     nb_chan = len(recording._channels)
        # else:
        # save binary file (chunk by hcunk) into a new file
        raw_filename = tmpdir / 'recording'
        n_chan = recording.get_num_channels()
        chunksize = 2**24 // n_chan
        dtype = 'int16'
        se.write_binary_dat_format(recording, raw_filename, time_axis=0, dtype=dtype, chunksize=chunksize)

        # save prb file:
        if p['probe_file'] is None:
            p['probe_file'] = tmpdir / 'probe.prb'
            se.save_probe_file(recording, p['probe_file'], format='klusta', radius=p['adjacency_radius'])

        if p['detect_sign'] < 0:
            detect_sign = 'negative'
        elif p['detect_sign'] > 0:
            detect_sign = 'positive'
        else:
            detect_sign = 'both'

        # set up klusta config file
        with (source_dir / 'config_default.prm').open('r') as f:
            klusta_config = f.readlines()

        # Note: should use format with dict approach here
        klusta_config = ''.join(klusta_config).format(raw_filename,
                                                      p['probe_file'], float(recording.get_sampling_frequency()),
                                                      recording.get_num_channels(), "'{}'".format(dtype),
                                                      p['threshold_strong_std_factor'], p['threshold_weak_std_factor'], "'" + detect_sign + "'",
                                                      p['extract_s_before'], p['extract_s_after'], p['n_features_per_channel'],
                                                      p['pca_n_waveforms_max'], p['num_starting_clusters']
                                                      )

        with (tmpdir / 'config.prm').open('w') as f:
            f.writelines(klusta_config)

        try:
            sorting = klusta_helper(tmpdir=tmpdir)
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)


def klusta_helper(*, tmpdir):  # all mlpr parameters
    cmd = 'klusta {} --overwrite'.format(tmpdir / 'config.prm')

    _call_command(cmd)
    if not (tmpdir / 'recording.kwik').is_file():
        raise Exception('Klusta did not run successfully')
    sorting = se.KlustaSortingExtractor(tmpdir / 'recording.kwik')
    return sorting


# To be shared across sorters (2019.05.05)
def _get_tmpdir(sorter_name):
    code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
    tmpdir0 = os.environ.get('TEMPDIR', '/tmp')
    tmpdir = os.path.join(tmpdir0, '{}-tmp-{}'.format(sorter_name, code))
    # reset the output folder
    if os.path.exists(tmpdir):
        shutil.rmtree(str(tmpdir))
    else:
        os.makedirs(tmpdir)
    return tmpdir


def _run_command_and_print_output(command):
    command_list = shlex.split(command, posix="win" not in sys.platform)
    with Popen(command_list, stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def _run_command_and_print_output_split(command_list):
    with Popen(command_list, stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def _call_command(command):
    command_list = shlex.split(command, posix="win" not in sys.platform)
    try:
        call(command_list)
    except CalledProcessError as e:
        raise Exception(e.output)


def _call_command_split(command_list):
    try:
        call(command_list)
    except CalledProcessError as e:
        raise Exception(e.output)



from matplotlib import pyplot as plt
from sklearn.decomposition import PCA
import numpy as np


class FeatureSpaceWidget:
    def __init__(self, *, recording, sorting, channels=None, unit_ids=None, width=14, height=7, snippet_len=100,
                 title='', max_num_spikes_per_unit=50):
        self._IX = recording
        self._OX = sorting
        self._channels = channels
        self._unit_ids = unit_ids
        self._width = width
        self._height = height
        self._figure = None
        self._snippet_len = snippet_len
        self._title = title
        self._max_num_spikes_per_unit = max_num_spikes_per_unit

    def plot(self):
        self._do_plot()

    def figure(self):
        return self._figure

    def _do_plot(self):
        units = self._unit_ids
        channels = self._channels
        if units is None:
            units = self._OX.get_unit_ids()
        channel_ids = self._IX.get_channel_ids()
        M = len(channel_ids)
        channel_locations = np.zeros((M, 2))
        for ii, ch in enumerate(channel_ids):
            loc = self._IX.get_channel_property(ch, 'location')
            channel_locations[ii, :] = loc[-2:]
        if channels is None:
            channels = channel_ids
        list = []
        for unit in units:
            st = self._OX.get_unit_spike_train(unit_id=unit)
            if st is not None:
                spikes = self._get_random_spike_waveforms(unit=unit, max_num=self._max_num_spikes_per_unit,
                                                          channels=channels)
                item = dict(
                    representative_waveforms=spikes,
                    title='Unit {}'.format(int(unit))
                )
                list.append(item)
            else:
                print(unit, ' spike train is None')
        print(np.shape(list[0]['representative_waveforms']))
        opts = {'channel': -1,
                'feature_extraction': 'PCA',
                'features_to_show': (2, 1)}
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            # self._plot_spike_shapes_multi(list,channel_locations=channel_locations[np.array(channels),:])
            #self._plot_spike_shapes_multi(list, channel_locations=None)
            self._plot_spikes_in_feature_space_multi(list, opts)

    def _plot_spikes_in_feature_space(self, spikes_dict, feature_space, opts):
        features_to_show = opts['features_to_show']
        waveforms = spikes_dict['representative_waveforms']
        features = feature_space.transform(waveforms.T)
        # TODO: Optionally exclude outliers
        plt.scatter(features[:, features_to_show[0]], features[:, features_to_show[1]],
                    label=spikes_dict['title'])

    def _plot_spikes_in_feature_space_multi(self, list, opts):
        channel = opts['channel']
        all_waveforms = None
        for spikes_dict in list:
            waveforms = spikes_dict['representative_waveforms']
            if channel == -1:
                s = np.shape(waveforms)
                spikes_dict['representative_waveforms'] = waveforms = waveforms.reshape(s[0] * s[1], s[2])
            else:
                spikes_dict['representative_waveforms'] = waveforms = waveforms[channel, :, :]
            if all_waveforms is None:
                all_waveforms = waveforms
            else:
                all_waveforms = np.concatenate([all_waveforms, waveforms], axis=1)
        feature_space = self._compute_features(all_waveforms, opts)

        f = plt.figure(figsize=(10, 10))
        for d in list:
            self._plot_spikes_in_feature_space(d, feature_space, opts)
        plt.legend()
        plt.xlabel('{} - {}'.format(opts['feature_extraction'], opts['features_to_show'][0]))
        plt.ylabel('{} - {}'.format(opts['feature_extraction'], opts['features_to_show'][1]))

    def _compute_features(self, spike_waveforms, opts):
        if opts['feature_extraction'] == 'PCA':
            features_obj = PCA()
            features_obj.fit(spike_waveforms.T)
        else:
            raise NotImplementedError('Override _compute_features if feature extraction other than PCA is required')
        return features_obj

    def _get_random_spike_waveforms(self, *, unit, max_num, channels):
        st = self._OX.get_unit_spike_train(unit_id=unit)
        num_events = len(st)
        if num_events > max_num:
            event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
        else:
            event_indices = range(num_events)

        spikes = self._IX.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=self._snippet_len,
                                       channel_ids=channels)
        if len(spikes) > 0:
            spikes = np.dstack(tuple(spikes))
        else:
            spikes = np.zeros((self._IX.get_num_channels(), self._snippet_len, 0))
        return spikes

    def _get_ylim_for_item(self, average_waveform=None, representative_waveforms=None):
        if average_waveform is None:
            if representative_waveforms is None:
                raise Exception('You must provide either average_waveform, representative waveforms, or both')
            average_waveform = np.mean(representative_waveforms, axis=2)
        return [average_waveform.min(), average_waveform.max()]

    def _determine_global_ylim(self, list):
        ret = [0, 0]
        for item in list:
            ylim0 = self._get_ylim_for_item(
                average_waveform=item.get('average_waveform', None),
                representative_waveforms=item.get('representative_waveforms', None)
            )
            ret[0] = np.minimum(ylim0[0], ret[0])
            ret[1] = np.maximum(ylim0[1], ret[1])
        return ret



import vdomr as vd
import time
import sys
import mtlogging
import numpy as np
import json
from .tablewidget import TableWidget


class AggregatedSortingResultsTableView(vd.Component):
    def __init__(self, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._aggregated_sorting_results = prepare_result.get('aggregated_sorting_results', None)
        self._table_widget = AggregatedSortingResultsTableWidget(context=context, aggregated_sorting_results=self._aggregated_sorting_results)

    # this will be done in a worker thread
    @staticmethod
    def prepareView(context, opts):
        # make sure it gets downloaded
        print('Loading aggregated sorting results...')
        aggregated_sorting_results = context.aggregatedSortingResults()

        return dict(
            aggregated_sorting_results=aggregated_sorting_results
        )

    def setSize(self, size):
        self._size = size
        if self._table_widget:
            self._table_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Aggregated results table'

    def render(self):
        return vd.div(
            self._table_widget
        )


class AggregatedSortingResultsTableWidget(vd.Component):
    def __init__(self, *, context, aggregated_sorting_results):
        vd.Component.__init__(self)
        self._size = (100, 100)
        self._context = context
        self._aggregated_sorting_results = aggregated_sorting_results
        self._asr = AggregatedSortingResults(self._aggregated_sorting_results)
        self._header_widget = _HeaderWidget(context, self._asr)
        self._header_widget.onChange(self._update_table)
        self._update_table()

    def setSize(self, size):
        self._size = size
        self._update_table()

    def _update_table(self):
        method = self._header_widget.method()
        accuracy_thresh = self._header_widget.accuracyThreshold()
        snr_thresh = self._header_widget.snrThreshold()
        study_names = self._asr.studyNames()
        sorter_names = self._asr.sorterNames()
        records = []
        for study_name in study_names:
            rec0 = dict(
                study_name=study_name
            )
            for sn in sorter_names:
                if method == 'acc_thr':
                    count0 = self._asr.getAccuracyCount(study_name=study_name, sorter_name=sn, accuracy_thresh=accuracy_thresh)
                    if count0 is not None:
                        val0 = str(count0)
                    else:
                        val0 = ''
                elif method == 'snr_thr':
                    avg0 = self._asr.getAverageAccuracy(study_name=study_name, sorter_name=sn, snr_thresh=snr_thresh)
                    if avg0 is not None:
                        val0 = str(int(avg0 * 1000) / 1000)
                    else:
                        val0 = ''
                else:
                    raise Exception('Unrecognized method: ' + method)
                rec0['sorter-' + sn] = val0
            records.append(rec0)
        columns0 = [
            dict(label=sorter_name, name='sorter-' + sorter_name)
            for sorter_name in sorter_names
        ]
        columns = [dict(label='Study', name='study_name')] + columns0

        self._table_widget = TableWidget(
            columns=columns,
            records=records,
            height=self._size[1] - self._header_widget.height()
        )
        self.refresh()

        # self._recording_ids = self._context.recordingIds()
        # recobjs = [self._context.recordingObject(recid) for recid in self._recording_ids]
        # records = [
        #     dict(
        #         study_name=obj['study'],
        #         recording_name=obj['name'],
        #         duration_sec=_get_computed_info(obj, 'duration_sec'),
        #         samplerate=_get_computed_info(obj, 'samplerate'),
        #         num_channels=_get_computed_info(obj, 'num_channels')
        #     )
        #     for obj in recobjs
        # ]

        # self._table_widget = TableWidget(
        #     columns = [
        #         dict(label='Study', name='study_name'),
        #         dict(label='Recording', name='recording_name'),
        #         dict(label='Duration (sec)', name='duration_sec'),
        #         dict(label='Sampling freq. (Hz)', name='samplerate'),
        #         dict(label='Num. channels', name='num_channels'),
        #     ],
        #     records = records,
        #     height=self._size[1]
        # )
        # self._table_widget.onSelectionChanged(self._on_widget_selection_changed)
        # self._on_context_selection_changed()
        # self.refresh()
    def _on_widget_selection_changed(self):
        pass
        # current_row_index = self._table_widget.currentRowIndex()
        # if current_row_index is not None:
        #     self._context.setCurrentRecordingId(self._recording_ids[current_row_index])
        # else:
        #     self._context.setCurrentRecordingId(None)

    def _on_context_selection_changed(self):
        pass
        # recid = self._context.currentRecordingId()
        # if recid is not None:
        #     try:
        #         index0 = self._recording_ids.index(recid)
        #     except:
        #         index0 = None
        # else:
        #     index0 = None
        # self._table_widget.setCurrentRowIndex(index0)

    def render(self):
        return vd.div(
            self._header_widget,
            self._table_widget
        )


class _HeaderWidget(vd.Component):
    def __init__(self, context, asr):
        vd.Component.__init__(self)
        self._asr = asr
        self._context = context

        self._method = 'snr_thr'

        self._acc_method_button = vd.components.RadioButton()
        self._acc_method_button.onChange(self._on_acc_method)
        self._accuracy_threshold = 0.8
        self._acc_thr_select_box = vd.components.SelectBox(style={'width': '100px'})
        self._acc_thr_select_box.onChange(self._on_selection_changed)

        self._snr_method_button = vd.components.RadioButton(checked=True)
        self._snr_method_button.onChange(self._on_snr_method)
        self._snr_threshold = 8
        self._snr_thr_select_box = vd.components.SelectBox(style={'width': '100px'})
        self._snr_thr_select_box.onChange(self._on_selection_changed)

        self._on_change_handlers = []
        self._update_options()

    def onChange(self, handler):
        self._on_change_handlers.append(handler)

    def _update_options(self):
        self._acc_thr_select_box.setOptions([str(num) for num in range(100, 0, -5)])
        self._acc_thr_select_box.setValue(str(int(self._accuracy_threshold * 100)))

        self._snr_thr_select_box.setOptions([str(num) for num in range(1, 100, 1)])
        self._snr_thr_select_box.setValue(str(int(self._snr_threshold)))

        self.refresh()

    def _on_acc_method(self):
        if self._acc_method_button.checked():
            self._snr_method_button.setChecked(False)
            self._method = 'acc_thr'
        else:
            return
        for handler in self._on_change_handlers:
            handler()

    def _on_snr_method(self):
        if self._snr_method_button.checked():
            self._acc_method_button.setChecked(False)
            self._method = 'snr_thr'
        else:
            return
        for handler in self._on_change_handlers:
            handler()

    def _on_selection_changed(self, value):
        val = self._acc_thr_select_box.value()
        self._accuracy_threshold = int(val) / 100

        val = self._snr_thr_select_box.value()
        self._snr_threshold = float(val)

        for handler in self._on_change_handlers:
            handler()

    def accuracyThreshold(self):
        return self._accuracy_threshold

    def snrThreshold(self):
        return self._snr_threshold

    def method(self):
        return self._method

    def height(self):
        return 40

    def render(self):
        return vd.div(
            vd.table(
                vd.tr(
                    vd.td(vd.div(self._snr_method_button, style={'margin-bottom': '6px'})),
                    vd.td(vd.div(style={'width': '5px'})),
                    vd.td('SNR threshold:'),
                    vd.td(self._snr_thr_select_box),

                    vd.td(vd.div(style={'width': '35px'})),

                    vd.td(vd.div(self._acc_method_button, style={'margin-bottom': '6px'})),
                    vd.td(vd.div(style={'width': '5px'})),
                    vd.td('Accuracy threshold:'),
                    vd.td(self._acc_thr_select_box)
                )
            ),
            style={'font-size': '12px'}
        )


class AggregatedSortingResults():
    def __init__(self, obj):
        self._obj = obj

        study_names = set()
        sorter_names = set()
        self._study_sorting_results_by_code = dict()
        for sr in obj.get('study_sorting_results', []):
            study0 = sr['study']
            sorter0 = sr['sorter']
            code0 = study0 + '----' + sorter0
            study_names.add(study0)
            sorter_names.add(sorter0)
            self._study_sorting_results_by_code[code0] = sr

        self._study_names = sorted(list(study_names))
        self._sorter_names = sorted(list(sorter_names))

    def studyNames(self):
        return self._study_names

    def sorterNames(self):
        return self._sorter_names

    def getStudySortingResult(self, study_name, sorter_name):
        code0 = study_name + '----' + sorter_name
        sr = self._study_sorting_results_by_code.get(code0, None)
        return sr

    def getAccuracyCount(self, study_name, sorter_name, accuracy_thresh):
        sr = self.getStudySortingResult(study_name=study_name, sorter_name=sorter_name)
        if sr is None:
            return None
        n1s = sr['num_false_positives']
        n2s = sr['num_matches']
        n3s = sr['num_false_negatives']
        snrs = sr['true_unit_snrs']
        accuracies = []
        for i in range(len(n1s)):
            tot0 = n1s[i] + n2s[i] + n3s[i]
            if tot0 == 0:
                acc0 = 0
            else:
                acc0 = n2s[i] / tot0
            accuracies.append(acc0)
        accuracies = np.array(accuracies)
        snrs = np.array(snrs)
        count0 = np.count_nonzero((accuracies >= accuracy_thresh))
        return count0

    def getAverageAccuracy(self, study_name, sorter_name, snr_thresh):
        sr = self.getStudySortingResult(study_name=study_name, sorter_name=sorter_name)
        if sr is None:
            return None
        n1s = sr['num_false_positives']
        n2s = sr['num_matches']
        n3s = sr['num_false_negatives']
        snrs = sr['true_unit_snrs']
        accuracies = []
        for i in range(len(n1s)):
            tot0 = n1s[i] + n2s[i] + n3s[i]
            if tot0 == 0:
                acc0 = 0
            else:
                acc0 = n2s[i] / tot0
            accuracies.append(acc0)
        accuracies = np.array(accuracies)
        snrs = np.array(snrs)
        inds = np.where(snrs >= snr_thresh)
        if inds is None:
            return None
        if len(inds[0]) == 0:
            return None
        avg0 = np.mean(accuracies[inds[0]])
        return avg0



from .extractors import SFMdaRecordingExtractor, SFMdaSortingExtractor, mdaio
from .extractors import EfficientAccessRecordingExtractor
from .extractors import loadProbeFile, saveProbeFile, writeBinaryDatFormat, getSubExtractorsByProperty
from .extractors import NeuroscopeSortingExtractor
from . import example_datasets



import vdomr as vd
import json


class SorterDefinitionsView(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)

    def setSize(self, size):
        self._size = size

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Sorter definitions'

    def render(self):
        # row info for the table
        rr = []
        sdkeys = self._context.sorterDefinitionKeys()
        for sdkey in sdkeys:
            sdef = self._context.sorterDefinition(sdkey)
            rr.append(dict(label=sdkey, value=vd.pre(json.dumps(sdef, indent=4))))

        rows = []
        for r in rr:
            rows.append(vd.tr(
                vd.th(r['label']), vd.td(r['value'])
            ))

        table = vd.table(rows, style={
                         'text-align': 'left', 'width': 'auto', 'font-size': '13px'}, class_='table')

        return vd.components.ScrollArea(vd.div(
            vd.h2('Sorter definitions for analysis: {}'.format(self._context.analysisName())),
            table, height=self._size[1]
        ))



import numpy as np
import struct
import os
import requests
import tempfile
import traceback


class MdaHeader:
    def __init__(self, dt0, dims0):
        uses64bitdims = (max(dims0) > 2e9)
        self.uses64bitdims = uses64bitdims
        self.dt_code = _dt_code_from_dt(dt0)
        self.dt = dt0
        self.num_bytes_per_entry = get_num_bytes_per_entry_from_dt(dt0)
        self.num_dims = len(dims0)
        self.dimprod = np.prod(dims0)
        self.dims = dims0
        if uses64bitdims:
            self.header_size = 3 * 4 + self.num_dims * 8
        else:
            self.header_size = (3 + self.num_dims) * 4

    def write(self, f):
        H = self
        _write_int32(f, H.dt_code)
        _write_int32(f, H.num_bytes_per_entry)
        if H.uses64bitdims:
            _write_int32(f, -H.num_dims)
            for j in range(0, H.num_dims):
                _write_int64(f, H.dims[j])
        else:
            _write_int32(f, H.num_dims)
            for j in range(0, H.num_dims):
                _write_int32(f, H.dims[j])


def npy_dtype_to_string(dt):
    str = dt.str[1:]
    map = {
        "f2": 'float16',
        "f4": 'float32',
        "f8": 'float64',
        "i1": 'int8',
        "i2": 'int16',
        "i4": 'int32',
        "u2": 'uint16',
        "u4": 'uint32'
    }
    return map[str]


class DiskReadMda:
    def __init__(self, path, header=None):
        self._npy_mode = False
        self._path = path
        if (file_extension(path) == '.npy'):
            raise Exception('DiskReadMda implementation has not been tested for npy files')
            # self._npy_mode = True
            # if header:
            #     raise Exception('header not allowed in npy mode for DiskReadMda')
        if header:
            self._header = header
            self._header.header_size = 0
        else:
            self._header = _read_header(self._path)

    def dims(self):
        if self._npy_mode:
            A = np.load(self._path, mmap_mode='r')
            return A.shape
        return self._header.dims

    def N1(self):
        return self.dims()[0]

    def N2(self):
        return self.dims()[1]

    def N3(self):
        return self.dims()[2]

    def dt(self):
        if self._npy_mode:
            A = np.load(self._path, mmap_mode='r')
            return npy_dtype_to_string(A.dtype)
        return self._header.dt

    def numBytesPerEntry(self):
        if self._npy_mode:
            A = np.load(self._path, mmap_mode='r')
            return A.itemsize
        return self._header.num_bytes_per_entry

    def readChunk(self, i1=-1, i2=-1, i3=-1, N1=1, N2=1, N3=1):
        # print("Reading chunk {} {} {} {} {} {}".format(i1,i2,i3,N1,N2,N3))
        if (i2 < 0):
            if self._npy_mode:
                A = np.load(self._path, mmap_mode='r')
                return A[:, :, i1:i1 + N1]
            return self._read_chunk_1d(i1, N1)
        elif (i3 < 0):
            if N1 != self.N1():
                print("Unable to support N1 {} != {}".format(N1, self.N1()))
                return None
            X = self._read_chunk_1d(i1 + N1 * i2, N1 * N2)
            if X is None:
                print('Problem reading chunk from file: ' + self._path)
                return None
            if self._npy_mode:
                A = np.load(self._path, mmap_mode='r')
                return A[:, i2:i2 + N2]
            return np.reshape(X, (N1, N2), order='F')
        else:
            if N1 != self.N1():
                print("Unable to support N1 {} != {}".format(N1, self.N1()))
                return None
            if N2 != self.N2():
                print("Unable to support N2 {} != {}".format(N2, self.N2()))
                return None
            if self._npy_mode:
                A = np.load(self._path, mmap_mode='r')
                return A[:, :, i3:i3 + N3]
            X = self._read_chunk_1d(i1 + N1 * i2 + N1 * N2 * i3, N1 * N2 * N3)
            return np.reshape(X, (N1, N2, N3), order='F')

    def _read_chunk_1d(self, i, N):
        offset = self._header.header_size + self._header.num_bytes_per_entry * i
        if is_url(self._path):
            tmp_fname = _download_bytes_to_tmpfile(self._path, offset, offset + self._header.num_bytes_per_entry * N)
            try:
                ret = self._read_chunk_1d_helper(tmp_fname, N, offset=0)
            except:
                ret = None
            # os.remove(tmp_fname)
            return ret
        return self._read_chunk_1d_helper(self._path, N, offset=offset)

    def _read_chunk_1d_helper(self, path0, N, *, offset):
        f = open(path0, "rb")
        try:
            f.seek(offset)
            ret = np.fromfile(f, dtype=self._header.dt, count=N)
            f.close()
            return ret
        except Exception as e:  # catch *all* exceptions
            print(e)
            f.close()
            return None


def is_url(path):
    path = path or ''
    return path.startswith('http://') or path.startswith('https://') or path.startswith(
        'kbucket://') or path.startswith('sha1://') or path.startswith('sha1dir://')


def _download_bytes_to_tmpfile(url, start, end):
    headers = {"Range": "bytes={}-{}".format(start, end - 1)}
    r = requests.get(url, headers=headers, stream=True)
    _, tmp_fname = tempfile.mkstemp()
    with open(tmp_fname, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
    return tmp_fname


def _read_header(path):
    if is_url(path):
        tmp_fname = _download_bytes_to_tmpfile(path, 0, 200)
        if not tmp_fname:
            raise Exception('Problem downloading bytes from ' + path)
        try:
            ret = _read_header(tmp_fname)
        except:
            ret = None
        os.remove(tmp_fname)
        return ret

    f = open(path, "rb")
    try:
        dt_code = _read_int32(f)
        _ = _read_int32(f)  # num bytes per entry
        num_dims = _read_int32(f)
        uses64bitdims = False
        if (num_dims < 0):
            uses64bitdims = True
            num_dims = -num_dims
        if (num_dims < 1) or (num_dims > 6):  # allow single dimension as of 12/6/17
            print("Invalid number of dimensions: {}".format(num_dims))
            f.close()
            return None
        dims = []
        dimprod = 1
        if uses64bitdims:
            for _ in range(0, num_dims):
                tmp0 = _read_int64(f)
                dimprod = dimprod * tmp0
                dims.append(tmp0)
        else:
            for _ in range(0, num_dims):
                tmp0 = _read_int32(f)
                dimprod = dimprod * tmp0
                dims.append(tmp0)
        dt = _dt_from_dt_code(dt_code)
        if dt is None:
            print("Invalid data type code: {}".format(dt_code))
            f.close()
            return None
        H = MdaHeader(dt, dims)
        if (uses64bitdims):
            H.uses64bitdims = True
            H.header_size = 3 * 4 + H.num_dims * 8
        f.close()
        return H
    except Exception as e:  # catch *all* exceptions
        print(e)
        f.close()
        return None


def _dt_from_dt_code(dt_code):
    if dt_code == -2:
        dt = 'uint8'
    elif dt_code == -3:
        dt = 'float32'
    elif dt_code == -4:
        dt = 'int16'
    elif dt_code == -5:
        dt = 'int32'
    elif dt_code == -6:
        dt = 'uint16'
    elif dt_code == -7:
        dt = 'float64'
    elif dt_code == -8:
        dt = 'uint32'
    else:
        dt = None
    return dt


def _dt_code_from_dt(dt):
    if dt == 'uint8':
        return -2
    if dt == 'float32':
        return -3
    if dt == 'int16':
        return -4
    if dt == 'int32':
        return -5
    if dt == 'uint16':
        return -6
    if dt == 'float64':
        return -7
    if dt == 'uint32':
        return -8
    return None


def get_num_bytes_per_entry_from_dt(dt):
    if dt == 'uint8':
        return 1
    if dt == 'float32':
        return 4
    if dt == 'int16':
        return 2
    if dt == 'int32':
        return 4
    if dt == 'uint16':
        return 2
    if dt == 'float64':
        return 8
    if dt == 'uint32':
        return 4
    return None


def readmda_header(path):
    if (file_extension(path) == '.npy'):
        raise Exception('Cannot read mda header for .npy file.')
    return _read_header(path)


def _write_header(path, H, rewrite=False):
    if rewrite:
        f = open(path, "r+b")
    else:
        f = open(path, "wb")
    try:
        H.write(f)
        f.close()
        return True
    except Exception as e:  # catch *all* exceptions
        print(e)
        f.close()
        return False


def readmda(path):
    if (file_extension(path) == '.npy'):
        return readnpy(path)
    H = _read_header(path)
    if (H is None):
        print("Problem reading header of: {}".format(path))
        return None
    ret = np.array([])
    f = open(path, "rb")
    try:
        f.seek(H.header_size)
        # This is how I do the column-major order
        ret = np.fromfile(f, dtype=H.dt, count=H.dimprod)
        ret = np.reshape(ret, H.dims, order='F')
        f.close()
        return ret
    except Exception as e:  # catch *all* exceptions
        print(e)
        f.close()
        return None


def writemda32(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy32(X, fname)
    return _writemda(X, fname, 'float32')


def writemda64(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy64(X, fname)
    return _writemda(X, fname, 'float64')


def writemda8(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy8(X, fname)
    return _writemda(X, fname, 'uint8')


def writemda32i(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy32i(X, fname)
    return _writemda(X, fname, 'int32')


def writemda32ui(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy32ui(X, fname)
    return _writemda(X, fname, 'uint32')


def writemda16i(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy16i(X, fname)
    return _writemda(X, fname, 'int16')


def writemda16ui(X, fname):
    if (file_extension(fname) == '.npy'):
        return writenpy16ui(X, fname)
    return _writemda(X, fname, 'uint16')


def writemda(X, fname, *, dtype):
    return _writemda(X, fname, dtype)


def _writemda(X, fname, dt):
    dt_code = 0
    # num_bytes_per_entry=get_num_bytes_per_entry_from_dt(dt)
    dt_code = _dt_code_from_dt(dt)
    if dt_code is None:
        print("Unexpected data type: {}".format(dt))
        return False

    if type(fname) == str:
        f = open(fname, 'wb')
    else:
        f = fname
    try:
        H = MdaHeader(dt0=dt, dims0=X.shape)
        H.write(f)

        # _write_int32(f,dt_code)
        # _write_int32(f,num_bytes_per_entry)
        # _write_int32(f,X.ndim)
        # for j in range(0,X.ndim):
        #     _write_int32(f,X.shape[j])

        # This is how I do column-major order
        # A=np.reshape(X,X.size,order='F').astype(dt)
        # A.tofile(f)

        bytes0 = X.astype(dt).tobytes(order='F')
        f.write(bytes0)

        if type(fname) == str:
            f.close()
        return True
    except Exception as e:  # catch *all* exceptions
        traceback.print_exc()
        print(e)
        if type(fname) == str:
            f.close()
        return False


def readnpy(path):
    return np.load(path)


def writenpy8(X, path):
    return _writenpy(X, path, dtype='int8')


def writenpy32(X, path):
    return _writenpy(X, path, dtype='float32')


def writenpy64(X, path):
    return _writenpy(X, path, dtype='float64')


def writenpy16i(X, path):
    return _writenpy(X, path, dtype='int16')


def writenpy16ui(X, path):
    return _writenpy(X, path, dtype='uint16')


def writenpy32i(X, path):
    return _writenpy(X, path, dtype='int32')


def writenpy32ui(X, path):
    return _writenpy(X, path, dtype='uint32')


def writenpy(X, path, *, dtype):
    return _writenpy(X, path, dtype=dtype)


def _writenpy(X, path, *, dtype):
    np.save(path, X.astype(dtype=dtype, copy=False))  # astype will always create copy if dtype does not match
    # apparently allowing pickling is a security issue. (according to the docs) ??
    # np.save(path,X.astype(dtype=dtype,copy=False),allow_pickle=False) # astype will always create copy if dtype does not match
    return True


def appendmda(X, path):
    if (file_extension(path) == '.npy'):
        raise Exception('appendmda not yet implemented for .npy files')
    H = _read_header(path)
    if (H is None):
        print("Problem reading header of: {}".format(path))
        return None
    if (len(H.dims) != len(X.shape)):
        print("Incompatible number of dimensions in appendmda", H.dims, X.shape)
        return None
    num_entries_old = np.product(H.dims)
    num_dims = len(H.dims)
    for j in range(num_dims - 1):
        if (X.shape[j] != X.shape[j]):
            print("Incompatible dimensions in appendmda", H.dims, X.shape)
            return None
    H.dims[num_dims - 1] = H.dims[num_dims - 1] + X.shape[num_dims - 1]
    try:
        _write_header(path, H, rewrite=True)
        f = open(path, "r+b")
        f.seek(H.header_size + H.num_bytes_per_entry * num_entries_old)
        A = np.reshape(X, X.size, order='F').astype(H.dt)
        A.tofile(f)
        f.close()
    except Exception as e:  # catch *all* exceptions
        print(e)
        f.close()
        return False


def file_extension(fname):
    if type(fname) == str:
        _, ext = os.path.splitext(fname)
        return ext
    else:
        return None


def _read_int32(f):
    return struct.unpack('<i', f.read(4))[0]


def _read_int64(f):
    return struct.unpack('<q', f.read(8))[0]


def _write_int32(f, val):
    f.write(struct.pack('<i', val))


def _write_int64(f, val):
    f.write(struct.pack('<q', val))


def _header_from_file(f):
    try:
        dt_code = _read_int32(f)
        _ = _read_int32(f)  # num bytes per entry
        num_dims = _read_int32(f)
        uses64bitdims = False
        if (num_dims < 0):
            uses64bitdims = True
            num_dims = -num_dims
        if (num_dims < 1) or (num_dims > 6):  # allow single dimension as of 12/6/17
            print("Invalid number of dimensions: {}".format(num_dims))
            return None
        dims = []
        dimprod = 1
        if uses64bitdims:
            for _ in range(0, num_dims):
                tmp0 = _read_int64(f)
                dimprod = dimprod * tmp0
                dims.append(tmp0)
        else:
            for _ in range(0, num_dims):
                tmp0 = _read_int32(f)
                dimprod = dimprod * tmp0
                dims.append(tmp0)
        dt = _dt_from_dt_code(dt_code)
        if dt is None:
            print("Invalid data type code: {}".format(dt_code))
            return None
        H = MdaHeader(dt, dims)
        if (uses64bitdims):
            H.uses64bitdims = True
            H.header_size = 3 * 4 + H.num_dims * 8
        return H
    except Exception as e:  # catch *all* exceptions
        print(e)
        return None


def mdaio_test():
    M = 4
    N = 12
    X = np.ndarray((M, N))
    for n in range(0, N):
        for m in range(0, M):
            X[m, n] = n * 10 + m
    writemda32(X, 'tmp1.mda')
    Y = readmda('tmp1.mda')
    print(Y)
    print(np.absolute(X - Y).max())
    Z = DiskReadMda('tmp1.mda')
    print(Z.readChunk(i1=0, i2=4, N1=M, N2=N - 4))

    # A = DiskWriteMda('tmpA.mda', (M, N))
    # A.writeChunk(Y, i1=0, i2=0)
    # B = readmda('tmpA.mda')
    # print(B.shape)
    # print(B)


# mdaio_test()



from .recordingsummaryview import RecordingSummaryView
from .timeseriesview import TimeseriesView
from .testplotlyview import TestPlotlyView
from .unitstableview import UnitsTableView
from .templatesview import TemplatesView
from .electrodegeometryview import ElectrodeGeometryView
from .currentstateview import CurrentStateView
from .unitdetailview import UnitDetailView


def get_recording_view_launchers():
    return [
        dict(
            name='recording-summary',
            label='Recording summary',
            view_class=RecordingSummaryView
        ),
        dict(
            name='electrode-geometry',
            label='Electrode geometry',
            view_class=ElectrodeGeometryView
        ),
        dict(
            name='timeseries',
            label='Timeseries',
            view_class=TimeseriesView
        ),
        dict(
            name='templates',
            label='Template plots',
            view_class=TemplatesView
        ),
        dict(
            name='true-units-table',
            label='Units table',
            view_class=UnitsTableView
        ),
        dict(
            name='current-unit-detail',
            label='Current unit detail',
            view_class=UnitDetailView
        ),
        dict(
            name='current-state',
            label='Current state',
            view_class=CurrentStateView
        )
    ]



from .sfmdaextractors import SFMdaRecordingExtractor, SFMdaSortingExtractor, mdaio
from .efficientaccess import EfficientAccessRecordingExtractor
from .neuroscopesortingextractor import NeuroscopeSortingExtractor
from .tools import loadProbeFile, saveProbeFile, writeBinaryDatFormat, getSubExtractorsByProperty



from spikeextractors import SortingExtractor
from spikeextractors.tools import read_python
import numpy as np
from pathlib import Path
import h5py


class KlustaSortingExtractor(SortingExtractor):
    def __init__(self, kwikfile):
        SortingExtractor.__init__(self)
        kwikfile = Path(kwikfile).absolute()
        F = h5py.File(kwikfile)
        channel_groups = F.get('channel_groups')
        self._spiketrains = []
        self._unit_ids = []
        for cgroup in channel_groups:
            group_id = int(cgroup)
            for cluster_id in channel_groups[cgroup]['clusters']['main']:
                clusters = np.array(channel_groups[cgroup]['spikes']['clusters']['main'])
                idx = np.nonzero(clusters == int(cluster_id))
                st = np.array(channel_groups[cgroup]['spikes']['time_samples'])[idx]
                self._spiketrains.append(st)
                self._unit_ids.append(int(cluster_id))
                self.set_unit_property(int(cluster_id), 'group', group_id)

    def get_unit_ids(self):
        return list(self._unit_ids)

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def write_sorting(sorting, save_path):
        save_path = Path(save_path)
        if save_path.is_dir():
            save_path = save_path / 'klusta.kwik'
        elif save_path.suffix == '.kwik':
            pass
        else:
            save_path.mkdir()
            save_path = save_path / 'klusta.kwik'
        F = h5py.File(save_path, 'w')
        F.attrs.create('kwik_version', data=2)
        if 'group' in sorting.get_unit_propertyNames():
            cgroups = np.unique([sorting.get_unit_property(unit, 'group') for unit in sorting.get_unit_ids()])
        else:
            cgroups = [0]

        channel_groups = F.create_group('channel_groups')

        for cgroup in cgroups:
            channel_group = channel_groups.create_group(str(cgroup))
            time_samples = np.array([])
            cluster_main = np.array([])
            if 'group' in sorting.get_unit_propertyNames():
                idxs = [unit for unit in sorting.get_unit_ids() if
                        sorting.get_unit_property(unit, 'group') == cgroup]
            else:
                idxs = sorting.get_unit_ids()
            clust = channel_group.create_group('clusters')
            clust.create_dataset('main', data=idxs)
            clust.create_dataset('original', data=idxs)
            for id in idxs:
                st = sorting.get_unit_spike_train(id)
                cl = [id] * len(sorting.get_unit_spike_train(id))
                time_samples = np.concatenate((time_samples, np.array(st)))
                cluster_main = np.concatenate((cluster_main, np.array(cl)))
            sorting_idxs = np.argsort(time_samples)
            time_samples = time_samples[sorting_idxs].astype(int)
            cluster_main = cluster_main[sorting_idxs].astype(int)
            spikes = channel_group.create_group('spikes')
            spikes.create_dataset('time_samples', data=time_samples)
            clusters = spikes.create_group('clusters')
            clusters.create_dataset('main', data=cluster_main)
            clusters.create_dataset('original', data=cluster_main)



import mlprocessors as mlpr
from mountaintools import client as mt
import os
from copy import deepcopy
import mtlogging


@mtlogging.log()
def summarize_sortings(sortings, compute_resource, label=None):
    print('')
    print('>>>>>> {}'.format(label or 'summarize sortings'))

    print('Gathering summarized sortings...')
    summarized_sortings = []
    for sorting in sortings:
        summary = dict()
        summary['plots'] = dict()

        sorting2 = deepcopy(sorting)
        sorting2['summary'] = summary
        summarized_sortings.append(sorting2)

    return summarized_sortings



import vdomr as vd
import time
import sys
import mtlogging
from mountaintools import client as mt
from .tablewidget import TableWidget
import json


class SortingResultsTableView(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._table_widget = SortingResultsTableWidget(
            context=self._context
        )
        self._table_widget.setSize(self._size)

    def setSize(self, size):
        self._size = size
        if self._table_widget:
            self._table_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Sorting results table'

    def render(self):
        return self._table_widget


class SortingResultsTableWidget(vd.Component):
    def __init__(self, *, context):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._sorting_result_names = self._context.sortingResultNames()
        self._context.onCurrentSortingResultChanged(self._on_context_selection_changed)
        self._update_table()

    def setSize(self, size):
        self._size = size
        self._update_table()

    def _update_table(self):
        self._table_widget = TableWidget(
            columns=[
                dict(label='Sorting result', name='sorting_result_name'),
                dict(label='Processor', name='processor_name'),
                dict(label='Parameters', name='parameters')
            ],
            records=[self._create_table_record(srname) for srname in self._sorting_result_names],
            height=self._size[1]
        )
        self._table_widget.onSelectionChanged(self._on_widget_selection_changed)
        self._on_context_selection_changed()
        self.refresh()

    def _create_table_record(self, srname):
        src = self._context.sortingResultContext(srname)
        sr = src.sortingResultObject()
        params = sr.get('sorter', {}).get('params', None)
        if params:
            params = _create_params_elmt(params)
        return dict(
            sorting_result_name=srname,
            processor_name=sr.get('sorter', {}).get('processor_name', None),
            parameters=params
        )

    def _on_widget_selection_changed(self):
        current_row_index = self._table_widget.currentRowIndex()
        if current_row_index is not None:
            self._context.setCurrentSortingResult(self._sorting_result_names[current_row_index])
        else:
            self._context.setCurrentSortingResult(None)

    def _on_context_selection_changed(self):
        srname = self._context.currentSortingResult()
        if srname is not None:
            try:
                index0 = self._sorting_result_names.index(srname)
            except:
                index0 = None
        else:
            index0 = None
        self._table_widget.setCurrentRowIndex(index0)

    def render(self):
        return self._table_widget


class SortingResultSelectComponent(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._widget = SortingResultSelectWidget(
            context=self._context
        )

    def render(self):
        return self._widget


class SortingResultSelectWidget(vd.Component):
    def __init__(self, *, context):
        vd.Component.__init__(self)
        self._context = context
        self._select_box = vd.components.SelectBox(style={'width': '100%', 'font-size': '12px'})
        self._select_box.onChange(self._on_selection_changed)
        self._context.onCurrentSortingResultChanged(self._on_context_selection_changed)
        self._update_options()

    def _update_options(self):
        self._sorting_result_names = self._context.sortingResultNames()
        self._select_box.setOptions(['[Select sorting result]'] + self._sorting_result_names)
        srname = self._context.currentSortingResult()
        if srname is not None:
            self._select_box.setValue(srname)
        self._on_context_selection_changed()
        self.refresh()

    def _on_selection_changed(self, value):
        index0 = self._select_box.index()
        if index0 > 0:
            srname = self._sorting_result_names[index0 - 1]
            self._context.setCurrentSortingResult(srname)

    def _on_context_selection_changed(self):
        srname = self._context.currentSortingResult()
        self._select_box.setValue(srname)

    def render(self):
        return self._select_box


def _create_params_elmt(params):
    items = []
    for key, val in params.items():
        items.append('{}={}'.format(key, json.dumps(val)))
    return '; '.join(items)
